'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
import os
import pandas as pd
from pm4py.objects.ocel.util import names_stripping
from enum import Enum
from pm4py.util import exec_utils, pandas_utils
from pm4py.objects.ocel.util import ocel_consistency
from pm4py.objects.ocel.util import filtering_utils
from pm4py.objects.ocel.exporter.util import clean_dataframes


class Parameters(Enum):
    ENABLE_NAMES_STRIPPING = "enable_names_stripping"


def apply(ocel: OCEL, file_path: str, parameters: Optional[Dict[Any, Any]] = None):
    """
    Exports the given OCEL (OCEL 2.0) into a SQLite database.
    Automatically converts the event and object 'ocel:timestamp' fields
    to Python datetime objects to avoid SQLite binding errors.
    """
    if parameters is None:
        parameters = {}

    enable_names_stripping = exec_utils.get_param_value(Parameters.ENABLE_NAMES_STRIPPING, parameters, True)

    import sqlite3

    # If the target file already exists, remove it to avoid conflicts
    if os.path.exists(file_path):
        os.remove(file_path)

    # Enforce consistency checks and propagate filtering
    ocel = ocel_consistency.apply(ocel, parameters=parameters)
    ocel = filtering_utils.propagate_relations_filtering(ocel, parameters=parameters)

    event_id = ocel.event_id_column
    event_activity = ocel.event_activity
    event_timestamp = ocel.event_timestamp
    object_id = ocel.object_id_column
    object_type = ocel.object_type_column
    qualifier = ocel.qualifier
    changed_field = ocel.changed_field

    conn = sqlite3.connect(file_path)

    # Write the "event" table
    EVENTS = ocel.events[[event_id, event_activity]].rename(
        columns={event_id: "ocel_id", event_activity: "ocel_type"}
    )
    EVENTS = EVENTS.drop_duplicates()
    EVENTS.to_sql("event", conn, index=False)

    # Write the "object" table
    OBJECTS = ocel.objects[[object_id, object_type]].rename(
        columns={object_id: "ocel_id", object_type: "ocel_type"}
    )
    OBJECTS = OBJECTS.drop_duplicates()
    OBJECTS.to_sql("object", conn, index=False)

    # Prepare event and object type mappings
    event_types = sorted(pandas_utils.format_unique(EVENTS["ocel_type"].unique()))
    object_types = sorted(pandas_utils.format_unique(OBJECTS["ocel_type"].unique()))

    EVENT_CORR_TYPE = pandas_utils.instantiate_dataframe(
        {"ocel_type": event_types, "ocel_type_map": event_types}
    )
    OBJECT_CORR_TYPE = pandas_utils.instantiate_dataframe(
        {"ocel_type": object_types, "ocel_type_map": object_types}
    )

    # Optionally strip names
    if enable_names_stripping:
        EVENT_CORR_TYPE["ocel_type_map"] = EVENT_CORR_TYPE["ocel_type_map"].apply(
            lambda x: names_stripping.apply(x)
        )
        OBJECT_CORR_TYPE["ocel_type_map"] = OBJECT_CORR_TYPE["ocel_type_map"].apply(
            lambda x: names_stripping.apply(x)
        )

    EVENT_CORR_TYPE.to_sql("event_map_type", conn, index=False)
    OBJECT_CORR_TYPE.to_sql("object_map_type", conn, index=False)

    # Write the event-object relationships
    E2O = ocel.relations[[event_id, object_id, qualifier]].rename(
        columns={event_id: "ocel_event_id", object_id: "ocel_object_id", qualifier: "ocel_qualifier"}
    )
    E2O.to_sql("event_object", conn, index=False)

    # Write the object-object relationships
    O2O = ocel.o2o.rename(
        columns={
            object_id: "ocel_source_id",
            object_id + "_2": "ocel_target_id",
            qualifier: "ocel_qualifier",
        }
    )
    O2O.to_sql("object_object", conn, index=False)

    # Export event tables by activity
    e_types = sorted(pandas_utils.format_unique(ocel.events[event_activity].unique()))
    for act in e_types:
        df = ocel.events[ocel.events[event_activity] == act].dropna(how="all", axis="columns")
        # Rename timestamp column
        df = df.rename(columns={event_id: "ocel_id", event_timestamp: "ocel_time"})

        # Force ID to string
        df["ocel_id"] = df["ocel_id"].astype("string")

        # Convert 'ocel_time' to Python datetime (if it's a pandas Timestamp or string)
        df["ocel_time"] = pd.to_datetime(df["ocel_time"], errors="coerce")
        for col in df.columns:
            if str(df[col].dtype) == "object":
                df[col] = df[col].map(clean_dataframes.normalize_value)

        act_red = names_stripping.apply(act) if enable_names_stripping else act

        df = df.drop_duplicates()
        df.to_sql("event_" + act_red, conn, index=False)

    # Export object tables by type
    o_types = sorted(pandas_utils.format_unique(ocel.objects[object_type].unique()))
    for ot in o_types:
        df = ocel.objects[ocel.objects[object_type] == ot].dropna(how="all", axis="columns")
        df = df.rename(columns={object_id: "ocel_id"})
        del df[object_type]

        # Object changes table
        df2 = ocel.object_changes[ocel.object_changes[object_type] == ot].dropna(how="all", axis="columns")
        if len(df2) > 0:
            del df2[object_type]
            df2 = df2.rename(
                columns={
                    object_id: "ocel_id",
                    event_timestamp: "ocel_time",
                    changed_field: "ocel_changed_field",
                }
            )

            # Convert 'ocel_time' to Python datetime (if it's a pandas Timestamp or string)
            df2["ocel_time"] = pd.to_datetime(df2["ocel_time"], errors="coerce")

            df = pandas_utils.concat([df, df2], axis=0)

        df["ocel_id"] = df["ocel_id"].astype("string")
        for col in df.columns:
            if str(df[col].dtype) == "object":
                df[col] = df[col].map(clean_dataframes.normalize_value)

        ot_red = names_stripping.apply(ot) if enable_names_stripping else ot

        df.to_sql("object_" + ot_red, conn, index=False)

    conn.close()



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import numpy as np
from typing import List, Union, Optional, Dict, Any
from pm4py.algo.transformation.to_embeddings.util import embed_sentence


def cosine_similarity(a, b):
    return np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))


def apply(target_embedding: Union[str, List[float]], embeddings_list: List[List[float]], parameters: Optional[Dict[Any, Any]] = None) -> List[float]:
    """
    Computes the cosine similarity between the embeddings of a target sentence and a list of embeddings
    (at the case or event level)

    Parameters
    -----------------
    target_embedding
        Target sentence (embeddings)
    embeddings_list
        List of embeddings (at the case or event level)
    parameters
        Parameters of the method

    Returns
    ------------------
    similarity_list
        List containing the cosine similarities between each entry and the target sentence
    """
    if parameters is None:
        parameters = {}

    if isinstance(target_embedding, str):
        target_embedding = embed_sentence.apply(target_embedding, parameters=parameters)

    similarity_list = [cosine_similarity(target_embedding, emb) for emb in embeddings_list]

    return similarity_list



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import importlib
import importlib.util
from typing import Any, Optional, Tuple


def _has_module(module_name: str) -> bool:
    try:
        return importlib.util.find_spec(module_name) is not None
    except ModuleNotFoundError:
        return False


def get_rustxes_backend_name() -> Optional[str]:
    if _has_module("r4pm.df"):
        return "r4pm"
    if _has_module("rustxes"):
        return "rustxes"
    return None


def import_rustxes_backend() -> Tuple[Any, str]:
    backend_name = get_rustxes_backend_name()
    if backend_name == "r4pm":
        return importlib.import_module("r4pm.df"), backend_name
    if backend_name == "rustxes":
        return importlib.import_module("rustxes"), backend_name
    raise ImportError("Neither `r4pm` nor `rustxes` is installed.")



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.util import exec_utils
from enum import Enum
from pm4py.algo.discovery.declare.variants import classic
from pm4py.objects.log.obj import EventLog
import pandas as pd
from typing import Union, Dict, Optional, Any


class Variants(Enum):
    CLASSIC = classic


def apply(
    log: Union[EventLog, pd.DataFrame],
    variant=Variants.CLASSIC,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Dict[str, Dict[Any, Dict[str, int]]]:
    """
    Discovers a DECLARE model from the provided event log

    Parameters
    ---------------
    log
        Log object (EventLog, Pandas dataframe)
    variant
        Variant of the algorithm to be used, including:
        - Variants.CLASSIC
    parameters
        Variant-specific parameters

    Returns
    ---------------
    declare_model
        DECLARE model (as Python dictionary), where each template is associated with its own rules
    """
    return exec_utils.get_variant(variant).apply(log, parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from typing import Optional, Dict, Any
from enum import Enum
from pm4py.util import exec_utils, pandas_utils
import pandas as pd


class Parameters(Enum):
    CONNECTION_STRING = "connection_string"


def apply(conn, parameters: Optional[Dict[Any, Any]] = None) -> pd.DataFrame:
    """
    Extracts an event log from the Camunda workflow system

    Parameters
    ---------------
    conn
        (if provided) ODBC connection object to the database (offering cursors)
    parameters
        Parameters of the algorithm, including:
        - Parameters.CONNECTION_STRING => connection string that is used (if no connection is provided)

    Returns
    ---------------
    dataframe
        Pandas dataframe
    """
    if parameters is None:
        parameters = {}

    import pm4py

    connection_string = exec_utils.get_param_value(
        Parameters.CONNECTION_STRING, parameters, None
    )

    if conn is None:
        import pyodbc

        conn = pyodbc.connect(connection_string)

    curs = conn.cursor()

    query = """
    SELECT
    pi.PROC_DEF_KEY_ AS "processID",
    ai.EXECUTION_ID_ AS "case:concept:name",
    ai.ACT_NAME_ AS "concept:name",
    ai.START_TIME_ AS "time:timestamp",
    ai.ASSIGNEE_ AS "org:resource"
    FROM
        act_hi_procinst pi
    JOIN
        act_hi_actinst ai ON pi.PROC_INST_ID_ = ai.PROC_INST_ID_
    ORDER BY
        pi.PROC_INST_ID_,
        ai.EXECUTION_ID_,
        ai.START_TIME_;
    """
    columns = [
        "processID",
        "case:concept:name",
        "concept:name",
        "time:timestamp",
        "org:resource",
    ]

    curs.execute(query)
    dataframe = curs.fetchall()
    dataframe = pandas_utils.instantiate_dataframe_from_records(
        dataframe, columns=columns
    )
    dataframe = pm4py.format_dataframe(dataframe)

    curs.close()
    conn.close()

    return dataframe



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util import exec_utils, constants, xes_constants
from typing import Optional, Dict, Any, List, Tuple
from pm4py.objects.conversion.log import converter as log_converter
import pandas as pd
from enum import Enum
from pm4py.algo.transformation.to_embeddings.util import similarity
from sentence_transformers import SentenceTransformer


class Parameters(Enum):
    EVENT_ID_KEY = "event_id_key"
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    EMBEDDING_MODEL = "embedding_model"
    ATTRIBUTE_KEY = constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY
    KEEP_CASES = "keep_cases"


def apply(log: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None) -> Tuple[List[str], List[List[float]]]:
    """
    Computes the embeddings of events considering a specific attribute

    Parameters
    -----------------
    log
        Pandas dataframe
    Parameters
        Parameters of the algorithm, including:
        - Parameters.EVENT_ID_KEY => an attribute unique per event
        - Parameters.EMBEDDING_MODEL => the embedding to be used (default: all-MiniLM-L6-v2)
        - Parameters.ATTRIBUTE_KEY => the attribute to be used

    Returns
    ----------------
    event_identifiers
        The list of all the event identifiers
    embeddings_list
        The list of embeddings for the considered events
    """
    if parameters is None:
        parameters = {}

    event_id_key = exec_utils.get_param_value(Parameters.EVENT_ID_KEY, parameters, constants.DEFAULT_INDEX_KEY)
    attribute_key = exec_utils.get_param_value(Parameters.ATTRIBUTE_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    embedding_model = exec_utils.get_param_value(Parameters.EMBEDDING_MODEL, parameters,
                                                 constants.DEFAULT_EMBEDDING_MODEL)

    log = log_converter.apply(log, variant=log_converter.Variants.TO_DATA_FRAME)

    event_identifiers = []
    sentences = []

    model = SentenceTransformer(embedding_model)

    log = log[[event_id_key, attribute_key]].to_dict("records")

    for entry in log:
        event_identifiers.append(entry[event_id_key])
        sentences.append(entry[attribute_key])

    embeddings_list = list(model.encode(sentences))

    return event_identifiers, embeddings_list


def keep_top_k_per_similarity(log: pd.DataFrame, target_sentence: str, k: int,
                              event_identifiers: Optional[List[str]] = None,
                              embeddings_list: Optional[List[List[float]]] = None,
                              parameters: Optional[Dict[Any, Any]] = None) -> pd.DataFrame:
    """
    Keeps the top K events for (embedding-based) similarity with the given sentence

    Parameters
    ----------------
    log
        Pandas dataframe
    target_sentence
        Target sentence
    k
        Number of similar events to retain
    event_identifiers
        (Optional) the list of event identifiers in the log, as returned by the 'apply' method
    embeddings_list
        (Optional) the list of embeddings for such events, as returned by the 'apply' method
    parameters
        Other parameters of the method:
        - Parameters.KEEP_CASES => If True, keep the cases containing such events, instead of the events themselves (default: False)

    Returns
    -----------------
    filtered_log
        Event log filtered on the top K events (or cases) according to the similarity metric.
    """
    if parameters is None:
        parameters = {}

    event_id_key = exec_utils.get_param_value(Parameters.EVENT_ID_KEY, parameters, constants.DEFAULT_INDEX_KEY)
    keep_cases = exec_utils.get_param_value(Parameters.KEEP_CASES, parameters, False)

    if event_identifiers is None or embeddings_list is None:
        event_identifiers, embeddings_list = apply(log, parameters=parameters)

    sim = similarity.apply(target_sentence, embeddings_list)
    events = [(event_identifiers[i], sim[i]) for i in range(len(event_identifiers))]
    events.sort(key=lambda x: (x[1], x[0]), reverse=True)
    events = events[:k]
    events = [x[0] for x in events]

    filt_log = log[log[event_id_key].isin(events)]

    if keep_cases:
        case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)
        filt_log2 = log[log[case_id_key].isin(filt_log[case_id_key])]
        return filt_log2

    return filt_log



"""
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
"""

from collections import defaultdict
from itertools import chain, combinations
from typing import Counter, Generic, TypeVar, Dict, Set, FrozenSet, Tuple, Iterator
from pm4py.objects.ocpn.obj import OCPetriNet, OCMarking
from pm4py.objects.ocpn.utils.ocpn_utils import pre_set

N = TypeVar("N", bound=OCPetriNet)
T = TypeVar("T", bound=OCPetriNet.Transition)
P = TypeVar("P", bound=OCPetriNet.Place)


class OCPetriNetSemantics(Generic[N]):

    @classmethod
    def is_enabled(cls, pn: N, transition: T, marking: OCMarking) -> bool:
        """
        Checks whether a given transition is enabled in a given object-centric Petri net and marking.
        A transition is enabled if every non-variable arc has at least one object in the corresponding place,
        and at least one arc can consume a token.

        Parameters
        ----------
        net
            object-centric Petri net
        transition
            transition to check
        marking
            marking to check

        Returns
        -------
        bool
            true if enabled, false otherwise
        """
        if transition not in pn.transitions:
            return False
        
        any_enabled = False
        for a in transition.in_arcs:
            if marking[a.source] == Counter():
                if not a.is_variable:
                    return False
            else:
                any_enabled = True
        return any_enabled
    
    @classmethod
    def is_binding_enabled(cls, pn: N, transition: T, marking: OCMarking, objects: Dict[str, Set]) -> bool:
        """
        Checks whether a given binding is enabled for a transition in a given object-centric Petri net and marking.

        Parameters
        ----------
        pn
            object-centric Petri net
        transition
            transition to check
        marking
            marking to check
        objects
            dict of objects per object type, e.g., {"order": {"order1", "order2"}}

        Returns
        -------
        bool
            true if binding is enabled, false otherwise
        """
        # We need to consume at least one token
        at_least_one_object = False
        
        # Check for every in-arc if:
        # - the binding respects the variability of the in-arc
        # - the state contains the necessary tokens
        for a in transition.in_arcs:
            ot = a.object_type
            
            obj_set = objects.get(ot, set())
            
            # Check if variability of arc is respected
            if (not a.is_variable) and (not len(obj_set) == 1):
                # nv-arc must be bound to exactly one object
                return False
            
            if not obj_set:
                # We assign not objects to this variable arc
                continue
            
            at_least_one_object = True
            
            # Check if state contains all tokens
            place_tokens = marking[a.source].keys()
            if not obj_set <= place_tokens:
                # tokens not present
                return False
        
        return at_least_one_object
    
    @classmethod
    def check_and_fire(
        cls, pn: N, transition: T, marking: OCMarking, objects: Dict[str, Set]
    ) -> OCMarking:
        """
        Checks if a transition is enabled and fires it with the given set of objects.
        If the transition is not enabled, it returns None.

        Parameters
        ----------
        pn
            object-centric Petri net
        transition
            transition to fire
        marking
            marking to use
        objects
            dict of objects per object type, e.g., {"order": {"order1", "order2"}}

        Returns
        -------
        OCMarking or None
            New marking after firing the transition or None if the transition was not enabled.
        """
        if cls.is_binding_enabled(pn, transition, marking, objects):
            return cls.fire(pn, transition, marking, objects)
        return None

    @classmethod
    def fire(
        cls, pn: N, transition: T, marking: OCMarking, objects: Dict[str, Set]
    ) -> OCMarking:
        """
        Execute a transition for a given set of objects per object type
        For performance reasons, this function does not check if the transition is enabled or binding passed set of objects is valid, i.e.,
        this should be performed by the invoking algorithm (if needed). Hence, markings can become negative.

        Parameters
        ----------
        pn
            object-centric Petri net
        transition
            transition to execute
        marking
            marking to use
        objects
            dict of objects per object type, e.g., {"order": {"order1", "order2"}}

        Returns
        -------
        OCMarking
            newly reached marking
        """
        # Copy counters place-wise so firing does not mutate the input marking.
        m_out = OCMarking({place: counter.copy() for place, counter in marking.items()})
        for a in transition.in_arcs:
            obj_set = objects.get(a.object_type, set())
            m_out[a.source] -= Counter(obj_set)
        for a in transition.out_arcs:
            obj_set = objects.get(a.object_type, set())
            m_out[a.target] += Counter(obj_set)
        return m_out
    
    @classmethod
    def replay(cls, ocpn: N, trace, initial_marking=None, final_marking=None):
        """
        Replays a trace on the object-centric Petri net.
        Starts with the initial marking and check if the trace reaches the final marking.
        
        Parameters
        ----------
        ocpn : N
            The object-centric Petri net to replay on.
        trace : tuple
            A trace from the object-centric Petri net,
            represented as a tuple of (transition_name, {object_type: set(object_ids)}) tuples.
        initial_marking : OCMarking, optional
            The initial marking to start the replay from. If None, the initial marking of the ocpn is used.
        final_marking : OCMarking, optional
            The final marking to check after the replay. If None, the final marking of the ocpn is used.
        """
        if not initial_marking:
            initial_marking = ocpn.initial_marking
        if not final_marking:
            final_marking = ocpn.final_marking
        
        transitions = {t.name: t for t in ocpn.transitions}
        
        # start in the initial marking
        marking = initial_marking
        
        # replay each binding
        for transition_name, objects in trace:
            t = transitions[transition_name]
            
            # Check if the binding is enabled
            if not cls.is_binding_enabled(ocpn, t, marking, objects):
                return False
            
            # Fire
            marking = cls.fire(ocpn, t, marking, objects)
        
        # Check if we are in the final marking
        return cls._is_final(marking, final_marking)

    @classmethod
    def _is_final(cls, marking: OCMarking, final_marking: OCMarking) -> bool:
        """
        Checks if the given marking is a valid final marking.
        May be overriden by subclasses.

        Parameters
        ----------
        marking
            marking to check
        final_marking
            final marking to check against

        Returns
        -------
        bool
            true if marking is a final marking, false otherwise
        """
        return marking == final_marking

    @classmethod
    def enabled_transitions(
        cls, pn: N, marking: OCMarking
    ) -> Set[T]:
        """
        Returns the enabled transitions in a given object-centric Petri net and marking

        Parameters
        ----------
        net
            object-centric Petri net
        marking
            marking to check

        Returns
        -------
        Set[T]
            Set of enabled transitions
        """
        return {t for t in pn.transitions if cls.is_enabled(pn, t, marking)}

    @classmethod
    def get_possible_bindings(
        cls, pn: N, transition: T, marking: OCMarking
    ) -> Iterator[Dict[str, Set]]:
        """
        Returns an Iterator on the possible bindings for a given transition 
        in a given object-centric Petri net and marking. 

        Parameters
        ----------
        net
            object-centric Petri net
        transition
            transition to check
        marking
            marking to check

        Returns
        -------
        Iterator[Dict[str, Set]]
            Iterator on all possible bindings (key is object type, value is a set of objects)
        """

        def get_common_objects(places: Set[P], marking: OCMarking) -> Set:
            """
            Get the intersection of objects in a set of places for the given marking.

            The resulting set contains objects present in ALL specified places.

            Parameters
            ----------
            places
                Set of places to consider
            marking
                Marking to consider

            Returns
            -------
            Set
                Set containing the common objects
            """
            if not places:
                return set()

            # iter to avoid treating the first place as a special case
            places_iter = iter(places)

            result_counter = marking[next(places_iter)].copy()

            for place in places_iter:
                if not result_counter:
                    break
                # intersect with the current place's objects
                result_counter &= marking[place]

            return set(result_counter)

        def get_powerset(object_type: str, objects: Set) -> Set[FrozenSet[Tuple]]:
            """
            Generate the powerset of a set of objects in the form (object_type, object_id).

            Parameters
            ----------
            object_type
                The type of the objects
            objects
                Set of objects to generate the powerset for

            Returns
            -------
            Set[FrozenSet[Tuple]]
                Set of frozen sets representing the powerset of the input set
            """
            items = [(object_type, obj_id) for obj_id in objects]
            powerset_iterator = chain.from_iterable(
                combinations(items, r) for r in range(len(items) + 1)
            )
            return {frozenset(subset) for subset in powerset_iterator}

        def recursive_bindings(
            variable_object_types: Set[str],
            non_variable_object_types: Set[str],
            available_objects: Dict[str, Set],
        ) -> Set[FrozenSet[Tuple]]:
            """
            Recursively generate all possible bindings given the remaining object types and available objects.

            Parameters
            ----------
            variable_object_types
                Set of variable object types to consider
            non_variable_object_types
                Set of non-variable object types to consider
            available_objects
                Dictionary where keys are object types and values are sets of available objects for that type

            Returns
            -------
            Set[FrozenSet[Tuple]]
                Set of possible bindings (as frozen sets of tuples (object_type, object_id))
            """
            # Base case
            if not non_variable_object_types and not variable_object_types:
                # empty binding
                return {frozenset()} 
        
            # Recursive case
            # Pick next object type to process. Non-variable object types are processed first.
            is_variable = not non_variable_object_types
            types_to_process = variable_object_types if is_variable else non_variable_object_types
            ot = types_to_process.pop()
            
            try:
                ot_objects = available_objects.get(ot, set())
                
                # Get all possible object combinations for this object type
                if is_variable:
                    # For variable object types, we can pick 0-all objects (powerset)
                    current_options = get_powerset(ot, ot_objects)
                else:
                    # For non-variable object types, we must pick exactly one object
                    current_options = {frozenset({(ot, obj)}) for obj in ot_objects}
                    # No options -> no bindings possible
                    if not current_options:
                        return set()
                    
                # Recursively get bindings for the remaining object types
                sub_bindings = recursive_bindings(
                    variable_object_types,
                    non_variable_object_types,
                    available_objects,
                )
                
                # cross product current options with sub-bindings
                bindings = {
                    sub_binding.union(option)
                    for option in current_options
                    for sub_binding in sub_bindings
                }
                
                return bindings
                
            finally:
                # restore the set of object types
                types_to_process.add(ot)


        if transition not in pn.transitions:
            # no binding to yield
            return

        # get object types involved in t and whether they are variable
        non_variable_object_types = set()
        variable_object_types = set()
        for a in transition.in_arcs:
            if a.is_variable:
                variable_object_types.add(a.object_type)
            else:
                non_variable_object_types.add(a.object_type)

        # get predecessor places per object type
        predecessors = {
            ot: pre_set(transition, ot)
            for ot in non_variable_object_types | variable_object_types
        }

        # per ot, get objects present in all predecessor places
        common_objects = {
            ot: get_common_objects(predecessors[ot], marking)
            for ot in non_variable_object_types | variable_object_types
        }

        # get bindings
        frozenset_bindings = recursive_bindings(
            variable_object_types, non_variable_object_types, common_objects
        )
        
        # Loop through immutable bindings and yield dictionary one by one
        for fs_binding in frozenset_bindings:
            # do not yield the empty binding
            if not fs_binding:
                continue
            
            binding_dict = defaultdict(set)
            for ot, obj in fs_binding:
                binding_dict[ot].add(obj)
                
            yield binding_dict



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from graphviz import Source
import tempfile

from pm4py.statistics.variants.log import get as variants_get
from pm4py.util import exec_utils, constants
from enum import Enum
from typing import Optional, Dict, Any
from pm4py.objects.log.obj import EventLog
from pm4py.util import typing
import graphviz


class Parameters(Enum):
    FORMAT = "format"
    ENABLE_GRAPH_TITLE = "enable_graph_title"
    GRAPH_TITLE = "graph_title"


def apply(
    log: EventLog,
    aligned_traces: typing.ListAlignments,
    parameters: Optional[Dict[Any, Any]] = None,
) -> graphviz.Source:
    """
    Gets the alignment table visualization from the alignments output

    Parameters
    -------------
    log
        Event log
    aligned_traces
        Aligned traces
    parameters
        Parameters of the algorithm

    Returns
    -------------
    gviz
        Graphviz object
    """
    if parameters is None:
        parameters = {}

    variants_idx_dict = variants_get.get_variants_from_log_trace_idx(
        log, parameters=parameters
    )

    variants_idx_list = []
    for variant in variants_idx_dict:
        variants_idx_list.append((variant, variants_idx_dict[variant]))
    # Deterministic ordering: sort by count (desc) then by variant key (asc)
    variants_idx_list = sorted(
        variants_idx_list, key=lambda x: (-len(x[1]), x[0])
    )

    image_format = exec_utils.get_param_value(
        Parameters.FORMAT, parameters, "png"
    )
    enable_graph_title = exec_utils.get_param_value(
        Parameters.ENABLE_GRAPH_TITLE,
        parameters,
        constants.DEFAULT_ENABLE_GRAPH_TITLES,
    )
    graph_title = exec_utils.get_param_value(
        Parameters.GRAPH_TITLE, parameters, "Alignments"
    )

    table_alignments_list = ["digraph {\n"]

    if enable_graph_title:
        table_alignments_list.append(
            'label=<<FONT POINT-SIZE="20">'
            + graph_title
            + '</FONT>>;\nlabelloc="top";\n'
        )

    table_alignments_list.append("tbl [\n")
    table_alignments_list.append("shape=plaintext\n")
    table_alignments_list.append("label=<\n")

    table_alignments_list.append(
        "<table border='0' cellborder='1' color='blue' cellspacing='0'>\n"
    )

    table_alignments_list.append(
        "<tr><td>Variant</td><td>Alignment</td></tr>\n"
    )

    for index, variant in enumerate(variants_idx_list):
        al_tr = aligned_traces[variant[1][0]]
        table_alignments_list.append("<tr>")
        table_alignments_list.append(
            "<td><font point-size='9'>Variant "
            + str(index + 1)
            + " ("
            + str(len(variant[1]))
            + " occurrences)</font></td>"
        )
        table_alignments_list.append(
            "<td><font point-size='6'><table border='0'><tr>"
        )
        for move in al_tr["alignment"]:
            if not (move[0] == ">>" or move[1] == ">>"):
                # sync move
                table_alignments_list.append(
                    '<td bgcolor="lightgreen">'
                    + str(move[1]).replace(">", "&gt;")
                    + "</td>"
                )
            elif move[1] == ">>":
                # move on log
                table_alignments_list.append(
                    '<td bgcolor="orange"><b>(LM)</b>'
                    + str(move[0]).replace(">", "&gt;")
                    + "</td>"
                )
            elif move[0] == ">>":
                # move on model
                table_alignments_list.append(
                    '<td bgcolor="violet"><b>(MM)</b>'
                    + str(move[1]).replace(">", "&gt;")
                    + "</td>"
                )
        table_alignments_list.append("</tr></table></font></td>")
        table_alignments_list.append("</tr>")

    table_alignments_list.append("</table>\n")
    table_alignments_list.append(">];\n")
    table_alignments_list.append("}\n")

    table_alignments = "".join(table_alignments_list)

    filename = tempfile.NamedTemporaryFile(suffix=".gv")
    filename.close()

    gviz = Source(table_alignments, filename=filename.name)

    gviz.format = image_format

    return gviz



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
"""Polars-based classic process cube variant."""

from bisect import bisect_right
import math
from enum import Enum
from typing import Optional, Dict, Any, Tuple, List
from itertools import product

import numpy as np
import polars as pl

from pm4py.util import constants, exec_utils

CASE_ID_COL = constants.CASE_CONCEPT_NAME

# Accepted aggregate functions for Polars pivot
_AGG_FN_ALIASES = {
    "avg": "mean",
    "average": "mean",
    "mean": "mean",
    "sum": "sum",
    "min": "min",
    "max": "max",
    "count": "count",
    "len": "count",
    "median": "median",
    "first": "first",
    "last": "last",
}


class Parameters(Enum):
    MAX_DIVISIONS_X = "max_divisions_x"
    MAX_DIVISIONS_Y = "max_divisions_y"
    AGGREGATION_FUNCTION = "aggregation_function"
    X_BINS = "x_bins"
    Y_BINS = "y_bins"


_NUMERIC_DTYPES = {
    pl.Int8,
    pl.Int16,
    pl.Int32,
    pl.Int64,
    pl.UInt8,
    pl.UInt16,
    pl.UInt32,
    pl.UInt64,
    pl.Float32,
    pl.Float64,
}


def _ensure_polars_df(feature_table: pl.LazyFrame | pl.DataFrame) -> pl.DataFrame:
    if isinstance(feature_table, pl.LazyFrame):
        return feature_table.collect()
    if isinstance(feature_table, pl.DataFrame):
        return feature_table.clone()
    raise TypeError(
        "feature_table must be a Polars LazyFrame or DataFrame"
    )


def _is_numeric_dtype(dtype: pl.DataType) -> bool:
    if dtype in _NUMERIC_DTYPES:
        return True
    dtype_str = str(dtype).lower()
    return any(token in dtype_str for token in ("int", "uint", "float"))


def _prepare_bins(
    series: pl.Series,
    manual_bins: Optional[List[float]],
    max_divisions: int,
) -> List[float]:
    if manual_bins:
        bins = sorted(set(float(b) for b in manual_bins))
        if len(bins) >= 2:
            return bins

    values = series.drop_nulls()
    if values.is_empty():
        return []

    min_val = float(values.min())
    max_val = float(values.max())
    if math.isfinite(min_val) and math.isfinite(max_val):
        if math.isclose(min_val, max_val):
            return [min_val - 0.5, max_val + 0.5]

        max_divisions = max(int(max_divisions), 1)
        bins_array = np.linspace(min_val, max_val, max_divisions + 1)
        bins = sorted(set(float(b) for b in bins_array))
        if len(bins) >= 2:
            return bins
        return [min_val - 0.5, max_val + 0.5]

    return []


def _bin_labels(bins: List[float]) -> List[str]:
    if len(bins) < 2:
        return []
    return [f"[{bins[i]}, {bins[i + 1]}]" for i in range(len(bins) - 1)]


def _assign_bins(series: pl.Series, bins: List[float], name: str) -> pl.Series:
    length = len(series)
    if len(bins) < 2:
        return pl.Series(name, [None] * length)

    labels = _bin_labels(bins)
    last_edge = bins[-1]

    assigned: List[Optional[str]] = []
    for value in series.to_list():
        if value is None:
            assigned.append(None)
            continue
        try:
            numeric_value = float(value)
        except (TypeError, ValueError):
            assigned.append(None)
            continue
        if not math.isfinite(numeric_value):
            assigned.append(None)
            continue

        idx = bisect_right(bins, numeric_value) - 1
        if idx == len(labels) and math.isclose(numeric_value, last_edge, rel_tol=1e-9, abs_tol=1e-9):
            idx -= 1
        if idx < 0 or idx >= len(labels):
            assigned.append(None)
        else:
            assigned.append(labels[idx])

    return pl.Series(name, assigned)


def _normalize_agg_fn(agg_fn: Any) -> str:
    if isinstance(agg_fn, str):
        key = agg_fn.lower()
        return _AGG_FN_ALIASES.get(key, key)
    return "mean"


def _aggregation_expression(column: str, agg_fn: str) -> pl.Expr:
    if agg_fn == "sum":
        return pl.col(column).sum()
    if agg_fn == "min":
        return pl.col(column).min()
    if agg_fn == "max":
        return pl.col(column).max()
    if agg_fn == "median":
        return pl.col(column).median()
    if agg_fn == "first":
        return pl.col(column).first()
    if agg_fn == "last":
        return pl.col(column).last()
    if agg_fn == "count":
        return pl.count()
    # default mean
    return pl.col(column).mean()


def _prefix_columns(df: pl.DataFrame, prefix: str) -> List[str]:
    return [col for col in df.columns if col.startswith(prefix + "_")]


def _normalize_dimension(dim: str | Tuple[str, ...]) -> Tuple[str, ...]:
    if isinstance(dim, tuple):
        return dim
    return (dim,)


def _select_bins_param(
    bins_param: Any,
    idx: int,
    col: str,
    num_dims: int,
) -> Optional[List[float]]:
    if bins_param is None:
        return None
    if isinstance(bins_param, dict):
        if col in bins_param:
            return bins_param[col]
        if idx in bins_param:
            return bins_param[idx]
        return None
    if isinstance(bins_param, (list, tuple)) and num_dims > 1:
        if len(bins_param) == num_dims and all(
            isinstance(item, (list, tuple)) for item in bins_param
        ):
            return list(bins_param[idx])
    return bins_param


def _combine_bins(dim_cols: Tuple[str, ...], bins_per_dim: List[List[str]]) -> List[str]:
    if not bins_per_dim:
        return []
    if len(bins_per_dim) == 1:
        return bins_per_dim[0]
    combined: List[str] = []
    for combo in product(*bins_per_dim):
        parts = [f"{dim_cols[i]}={combo[i]}" for i in range(len(dim_cols))]
        combined.append(" | ".join(parts))
    return combined


def _dimension_bins(
    df: pl.DataFrame,
    dim: str | Tuple[str, ...],
    bins_param: Any,
    max_divisions: int,
    dim_name: str,
) -> Tuple[pl.DataFrame, List[str]]:
    dim_cols = _normalize_dimension(dim)
    if not dim_cols:
        return pl.DataFrame(), []

    per_attr_dfs: List[pl.DataFrame] = []
    bins_per_dim: List[List[str]] = []

    for idx, col in enumerate(dim_cols):
        col_exists = col in df.columns
        numeric = col_exists and _is_numeric_dtype(df.schema[col])
        bin_col = f"__bin_{idx}"

        if numeric:
            selected_bins = _select_bins_param(bins_param, idx, col, len(dim_cols))
            bins = _prepare_bins(df[col], selected_bins, max_divisions)
            if len(bins) < 2:
                return pl.DataFrame(), []
            bins_per_dim.append(_bin_labels(bins))

            series = _assign_bins(df[col], bins, bin_col)
            attr_df = (
                df.select([CASE_ID_COL])
                .with_columns(series)
                .filter(pl.col(bin_col).is_not_null())
                .select([CASE_ID_COL, pl.col(bin_col)])
            )
        else:
            prefix_cols = _prefix_columns(df, col)
            if prefix_cols:
                bins_per_dim.append(prefix_cols)
                attr_df = (
                    df.select([CASE_ID_COL, *prefix_cols])
                    .melt(
                        id_vars=[CASE_ID_COL],
                        value_vars=prefix_cols,
                        variable_name=bin_col,
                        value_name="__value",
                    )
                    .filter(pl.col("__value").fill_null(0) >= 1)
                    .select([CASE_ID_COL, pl.col(bin_col)])
                )
            elif col_exists:
                series = df[col].cast(pl.Utf8)
                bins = (
                    series.drop_nulls()
                    .unique()
                    .sort()
                    .to_list()
                )
                if not bins:
                    return pl.DataFrame(), []
                bins_per_dim.append([str(b) for b in bins])
                attr_df = (
                    df.select([CASE_ID_COL])
                    .with_columns(series.alias(bin_col))
                    .filter(pl.col(bin_col).is_not_null())
                    .select([CASE_ID_COL, pl.col(bin_col)])
                )
            else:
                return pl.DataFrame(), []

        if attr_df.is_empty():
            return pl.DataFrame(), []
        per_attr_dfs.append(attr_df)

    combined = per_attr_dfs[0]
    for attr_df in per_attr_dfs[1:]:
        combined = combined.join(attr_df, on=CASE_ID_COL, how="inner")
        if combined.is_empty():
            return pl.DataFrame(), []

    bin_cols = [f"__bin_{idx}" for idx in range(len(dim_cols))]
    if len(bin_cols) == 1:
        combined = combined.with_columns(
            pl.col(bin_cols[0]).alias(f"{dim_name}_bin")
        )
        all_bins = bins_per_dim[0]
    else:
        label_parts = [
            pl.concat_str([pl.lit(f"{dim_cols[i]}="), pl.col(bin_cols[i])], separator="")
            for i in range(len(dim_cols))
        ]
        combined = combined.with_columns(
            pl.concat_str(label_parts, separator=" | ").alias(f"{dim_name}_bin")
        )
        all_bins = _combine_bins(dim_cols, bins_per_dim)

    return combined.select([CASE_ID_COL, f"{dim_name}_bin"]), all_bins


def _numeric_numeric_case(
    df: pl.DataFrame,
    x_col: str,
    y_col: str,
    agg_col: str,
    x_bins: List[float],
    y_bins: List[float],
) -> pl.DataFrame:
    x_bin_series = _assign_bins(df[x_col], x_bins, "x_bin")
    y_bin_series = _assign_bins(df[y_col], y_bins, "y_bin")

    df = df.with_columns([x_bin_series, y_bin_series])

    temp_df = (
        df.filter(pl.col("x_bin").is_not_null() & pl.col("y_bin").is_not_null())
        .select([pl.col(CASE_ID_COL), pl.col("x_bin"), pl.col("y_bin"), pl.col(agg_col)])
    )
    return temp_df


def _numeric_prefix_case(
    df: pl.DataFrame,
    x_col: str,
    agg_col: str,
    bins: List[float],
    prefix_cols: List[str],
    prefix_label: str,
) -> pl.DataFrame:
    x_bin_series = _assign_bins(df[x_col], bins, "x_bin")
    df = df.with_columns(x_bin_series)

    if not prefix_cols:
        return pl.DataFrame({CASE_ID_COL: [], "x_bin": [], "y_bin": [], agg_col: []})

    melted = (
        df.select([CASE_ID_COL, agg_col, "x_bin", *prefix_cols])
        .melt(
            id_vars=[CASE_ID_COL, agg_col, "x_bin"],
            value_vars=prefix_cols,
            variable_name=prefix_label,
            value_name="__value",
        )
        .filter(pl.col("x_bin").is_not_null() & (pl.col("__value").fill_null(0) >= 1))
        .select([
            pl.col(CASE_ID_COL),
            pl.col("x_bin"),
            pl.col(prefix_label).alias("y_bin"),
            pl.col(agg_col),
        ])
    )
    return melted


def _prefix_numeric_case(
    df: pl.DataFrame,
    y_col: str,
    agg_col: str,
    bins: List[float],
    prefix_cols: List[str],
    prefix_label: str,
) -> pl.DataFrame:
    y_bin_series = _assign_bins(df[y_col], bins, "y_bin")
    df = df.with_columns(y_bin_series)

    if not prefix_cols:
        return pl.DataFrame({CASE_ID_COL: [], "x_bin": [], "y_bin": [], agg_col: []})

    melted = (
        df.select([CASE_ID_COL, agg_col, "y_bin", *prefix_cols])
        .melt(
            id_vars=[CASE_ID_COL, agg_col, "y_bin"],
            value_vars=prefix_cols,
            variable_name=prefix_label,
            value_name="__value",
        )
        .filter(pl.col("y_bin").is_not_null() & (pl.col("__value").fill_null(0) >= 1))
        .select([
            pl.col(CASE_ID_COL),
            pl.col(prefix_label).alias("x_bin"),
            pl.col("y_bin"),
            pl.col(agg_col),
        ])
    )
    return melted


def _prefix_prefix_case(
    df: pl.DataFrame,
    agg_col: str,
    x_prefix_cols: List[str],
    y_prefix_cols: List[str],
) -> pl.DataFrame:
    if not x_prefix_cols or not y_prefix_cols:
        return pl.DataFrame({CASE_ID_COL: [], "x_bin": [], "y_bin": [], agg_col: []})

    x_melt = (
        df.select([CASE_ID_COL, agg_col, *x_prefix_cols])
        .melt(
            id_vars=[CASE_ID_COL, agg_col],
            value_vars=x_prefix_cols,
            variable_name="x_bin",
            value_name="__x_value",
        )
        .filter(pl.col("__x_value").fill_null(0) >= 1)
        .select([pl.col(CASE_ID_COL), pl.col("x_bin"), pl.col(agg_col)])
    )

    y_melt = (
        df.select([CASE_ID_COL, *y_prefix_cols])
        .melt(
            id_vars=[CASE_ID_COL],
            value_vars=y_prefix_cols,
            variable_name="y_bin",
            value_name="__y_value",
        )
        .filter(pl.col("__y_value").fill_null(0) >= 1)
        .select([pl.col(CASE_ID_COL), pl.col("y_bin")])
    )

    temp_df = (
        x_melt.join(y_melt, on=CASE_ID_COL, how="inner")
        .select([pl.col(CASE_ID_COL), pl.col("x_bin"), pl.col("y_bin"), pl.col(agg_col)])
    )
    return temp_df


def apply(
    feature_table: pl.LazyFrame | pl.DataFrame,
    x_col: str | Tuple[str, ...],
    y_col: str | Tuple[str, ...],
    agg_col: str,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Tuple[pl.DataFrame, Dict[Any, Any]]:
    """Construct a process cube using Polars data structures.

    The X/Y dimensions can be defined by a single attribute (str) or by a tuple of
    attributes. When a tuple is provided, bins are computed per attribute and then
    combined into a single composite bin label.
    """

    if parameters is None:
        parameters = {}

    df = _ensure_polars_df(feature_table)
    if df.is_empty():
        return pl.DataFrame(), {}

    agg_fn_param = exec_utils.get_param_value(
        Parameters.AGGREGATION_FUNCTION, parameters, "mean"
    )
    agg_fn = _normalize_agg_fn(agg_fn_param)

    max_divisions_x = exec_utils.get_param_value(Parameters.MAX_DIVISIONS_X, parameters, 4)
    max_divisions_y = exec_utils.get_param_value(Parameters.MAX_DIVISIONS_Y, parameters, 4)
    x_bins_param = exec_utils.get_param_value(Parameters.X_BINS, parameters, None)
    y_bins_param = exec_utils.get_param_value(Parameters.Y_BINS, parameters, None)

    x_dim_df, all_x_bins = _dimension_bins(
        df, x_col, x_bins_param, max_divisions_x, "x"
    )
    if x_dim_df.is_empty():
        return pl.DataFrame(), {}

    y_dim_df, all_y_bins = _dimension_bins(
        df, y_col, y_bins_param, max_divisions_y, "y"
    )
    if y_dim_df.is_empty():
        return pl.DataFrame(), {}

    temp_df = (
        x_dim_df.join(y_dim_df, on=CASE_ID_COL, how="inner")
        .join(df.select([CASE_ID_COL, agg_col]), on=CASE_ID_COL, how="inner")
    )

    if temp_df.is_empty():
        return pl.DataFrame(), {}

    temp_df = temp_df.filter(
        pl.col("x_bin").is_not_null() & pl.col("y_bin").is_not_null()
    )
    if temp_df.is_empty():
        return pl.DataFrame(), {}

    agg_expr = _aggregation_expression(agg_col, agg_fn).alias("value")
    grouped_values = (
        temp_df.group_by(["x_bin", "y_bin"], maintain_order=True)
        .agg(agg_expr)
    )

    pivot_df = grouped_values.pivot(
        values="value",
        index="x_bin",
        columns="y_bin",
        sort_columns=False,
    )

    if all_x_bins:
        base_rows = pl.DataFrame({"x_bin": all_x_bins})
        pivot_df = base_rows.join(pivot_df, on="x_bin", how="left")

    value_columns = [col for col in pivot_df.columns if col != "x_bin"]
    if all_y_bins:
        missing_cols = [col for col in all_y_bins if col not in value_columns]
        if missing_cols:
            additions = {col: [None] * pivot_df.height for col in missing_cols}
            pivot_df = pl.concat([pivot_df, pl.DataFrame(additions)], how="horizontal")
        ordered_cols = ["x_bin", *all_y_bins]
        remaining_cols = [col for col in pivot_df.columns if col not in ordered_cols]
        pivot_df = pivot_df.select([*ordered_cols, *remaining_cols])

    case_groups = (
        temp_df.group_by(["x_bin", "y_bin"], maintain_order=True)
        .agg(pl.col(CASE_ID_COL).unique().alias("case_ids"))
    )
    cell_case_dict = {
        (row[0], row[1]): set(row[2])
        for row in case_groups.iter_rows()
    }

    return pivot_df, cell_case_dict


__all__ = ["Parameters", "apply"]



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from copy import copy
from enum import Enum
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import pandas as pd
from pm4py.algo.conformance.alignments.petri_net import algorithm as petri_alignments
from pm4py.objects.conversion.log import converter as log_converter
from pm4py.objects.log.obj import EventLog, EventStream
from pm4py.objects.petri_net import semantics
from pm4py.objects.petri_net.obj import Marking, PetriNet
from pm4py.objects.petri_net.utils import align_utils as pn_align_utils, check_soundness
from pm4py.statistics.variants.log import get as variants_get
from pm4py.util import constants, exec_utils


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    TOKEN_REPLAY_VARIANT = "token_replay_variant"
    CLEANING_TOKEN_FLOOD = "cleaning_token_flood"
    SHOW_PROGRESS_BAR = "show_progress_bar"
    MULTIPROCESSING = "multiprocessing"
    CORES = "cores"

    # Optional: if True, uses transition *names* (task IDs) as symbols instead of labels.
    # Useful to disambiguate duplicate labels. Still ignores invisible transitions.
    USE_TASK_IDS = "use_task_ids"


SKIP = ">>"
Prefix = Tuple[str, ...]


def _get_alignment_model_part(step: Any) -> Tuple[Any, Any]:
    """
    Extract (model_transition_name, model_label) from one alignment step.

    Expected format when PM4Py is called with ret_tuple_as_trans_desc=True:
        step == ( (log_trans_name, model_trans_name), (log_label, model_label) )
    where the first tuple comes from the sync-product transition name, and the second from its label.

    Returns
    -------
    model_trans_name, model_label
    """
    try:
        trans_desc, labels = step
        model_trans_name = trans_desc[1]
        model_label = labels[1]
        return model_trans_name, model_label
    except Exception as exc:
        raise ValueError(
            "Unexpected alignment step format. Make sure alignments are computed with "
            "parameters={'ret_tuple_as_trans_desc': True}."
        ) from exc


def _extract_model_sequence(
    alignment: List[Any], *, use_task_ids: bool = False
) -> List[str]:
    """
    Implements the paper's λ̄(γ): projection of *model moves* of an alignment γ,
    yielding a complete activity sequence of the model.

    - Log-only moves are ignored (model side '>>')
    - Invisible model moves are ignored (model label is None)
    - If use_task_ids=True: return model transition *names* (task IDs) for visible moves
      (still ignoring invisible transitions).
    """
    seq: List[str] = []
    for step in alignment:
        model_trans_name, model_label = _get_alignment_model_part(step)

        if model_trans_name == SKIP:
            continue

        if model_label is None or model_label == SKIP:
            continue

        if use_task_ids:
            seq.append(str(model_trans_name))
        else:
            seq.append(str(model_label))

    return seq


def _update_prefix_stats(
    seq: List[str],
    weight: int,
    next_by_prefix: Dict[Prefix, Set[str]],
    prefix_weight: Dict[Prefix, int],
) -> None:
    """
    For each prefix of seq (excluding the full seq), update:
      - next_by_prefix[prefix] : set(next symbols)
      - prefix_weight[prefix]  : ω(prefix) aggregated across traces/variants

    This corresponds to the paper's ω(s) for the Λ1 setting:
      ω(γ) = frequency(trace)
      ω(s) = Σ ω(γ) for all alignments γ whose λ̄(γ) has prefix s.
    """
    if len(seq) < 2:
        return

    prefix: Prefix = ()
    for i in range(len(seq) - 1):
        prefix = prefix + (seq[i],)
        next_sym = seq[i + 1]
        next_by_prefix.setdefault(prefix, set()).add(next_sym)
        prefix_weight[prefix] = prefix_weight.get(prefix, 0) + weight


def apply(
    log: Union[EventLog, EventStream, pd.DataFrame],
    net: PetriNet,
    im: Marking,
    fm: Marking,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> float:
    """
    Alignment-based ETC precision (per "Measuring precision of modeled behavior").

    Summary of approach (Λ1 case):
      1) compute one optimal alignment per trace variant
      2) build the alignment automaton from λ̄(γ) (projection of model moves)
      3) compute precision from escaping edges:
           precision = 1 - (Σ ω(s)*|a_v(s) \\ e_x(s)|) / (Σ ω(s)*|a_v(s)|)

    Fixes vs the original snippet:
      - empty-prefix executed set taken from aligned sequences (not raw log)
      - available actions aggregated over all markings reached by the same prefix
      - tuple prefixes instead of comma-joined strings
      - robust log conversion and trace counting
    """
    if parameters is None:
        parameters = {}

    if not check_soundness.check_easy_soundness_net_in_fin_marking(net, im, fm):
        raise ValueError(
            "Align-ETC precision can only be applied on a Petri net that is "
            "a sound WF-net (easy sound)."
        )

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, "concept:name"
    )
    debug_level = parameters.get("debug_level", 0)
    use_task_ids = bool(exec_utils.get_param_value(Parameters.USE_TASK_IDS, parameters, False))

    # Convert input to EventLog for consistent handling
    ev_log: EventLog = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    # 1) Variants -> align fewer traces (one per variant)
    variants = variants_get.get_variants(
        ev_log, parameters={constants.PARAMETER_CONSTANT_ACTIVITY_KEY: activity_key}
    )
    variant_keys = list(variants.keys())

    red_log = EventLog()
    variant_freqs: List[int] = []
    for vkey in variant_keys:
        traces = variants[vkey]
        if not traces:
            continue
        red_log.append(traces[0])
        variant_freqs.append(len(traces))

    # 2) Align each variant trace
    align_params = dict(parameters)
    align_params["ret_tuple_as_trans_desc"] = True
    aligned_variants = petri_alignments.apply(red_log, net, im, fm, parameters=align_params)

    trans_by_name = {t.name: t for t in net.transitions}

    # 3) Build automaton stats from aligned model projections λ̄(γ)
    next_by_prefix: Dict[Prefix, Set[str]] = {}
    prefix_weight: Dict[Prefix, int] = {}
    start_actions: Set[str] = set()

    # Collect all markings reached at each prefix across all aligned traces
    prefix_markings: Dict[Prefix, Set[Marking]] = {}

    for idx, aligned in enumerate(aligned_variants):
        alignment = aligned["alignment"]
        freq = variant_freqs[idx]

        seq = _extract_model_sequence(alignment, use_task_ids=use_task_ids)

        if seq:
            start_actions.add(seq[0])

        _update_prefix_stats(seq, freq, next_by_prefix, prefix_weight)

        if len(seq) < 2:
            continue

        max_prefix_len = len(seq) - 1  # exclude last symbol

        marking = copy(im)
        prefix: Prefix = ()
        seen_visible = 0

        for step in alignment:
            model_trans_name, model_label = _get_alignment_model_part(step)

            # Execute any model-side transition (sync or model-only, incl. invisible)
            if model_trans_name != SKIP:
                if model_trans_name not in trans_by_name:
                    raise KeyError(
                        f"Transition name {model_trans_name!r} from alignment not found in model."
                    )
                new_marking = semantics.execute(trans_by_name[model_trans_name], net, marking)
                if new_marking is None:
                    raise RuntimeError(
                        f"Alignment tried to fire transition {model_trans_name!r} "
                        f"which is not enabled in the current marking."
                    )
                marking = new_marking

            # Update prefix only for visible model moves
            if model_trans_name == SKIP:
                continue
            if model_label is None or model_label == SKIP:
                continue

            seen_visible += 1
            sym = str(model_trans_name) if use_task_ids else str(model_label)
            prefix = prefix + (sym,)

            if seen_visible <= max_prefix_len:
                prefix_markings.setdefault(prefix, set()).add(marking)
            else:
                break

    # 4) Compute available actions a_v(s) and escaping edges
    enabled_cache: Dict[Marking, Set[str]] = {}

    def enabled_symbols_at(m: Marking) -> Set[str]:
        if m in enabled_cache:
            return enabled_cache[m]
        vis_trans = pn_align_utils.get_visible_transitions_eventually_enabled_by_marking(net, m)
        if use_task_ids:
            enabled = {str(t.name) for t in vis_trans}
        else:
            enabled = {str(t.label) for t in vis_trans if t.label is not None}
        enabled_cache[m] = enabled
        return enabled

    sum_at = 0.0
    sum_ee = 0.0

    for prefix, markings in prefix_markings.items():
        if prefix not in prefix_weight:
            continue

        available: Set[str] = set()
        for m in markings:
            available |= enabled_symbols_at(m)

        executed = next_by_prefix.get(prefix, set())
        escaping = available.difference(executed)

        w = prefix_weight[prefix]
        sum_at += len(available) * w
        sum_ee += len(escaping) * w

    # 5) Empty prefix ε
    n_traces = len(ev_log)
    enabled_ini = enabled_symbols_at(im)
    escaping_ini = enabled_ini.difference(start_actions)

    sum_at += len(enabled_ini) * n_traces
    sum_ee += len(escaping_ini) * n_traces

    precision = 1.0
    if sum_at > 0:
        precision = 1.0 - float(sum_ee) / float(sum_at)

    if debug_level:
        print(
            "[Align-ETC Precision (aligned automaton)] "
            f"AT={sum_at:.0f} EE={sum_ee:.0f} precision={precision:.5f}"
        )

    return precision



"""
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
"""

from pm4py.objects.oc_causal_net.obj import OCCausalNet
from pm4py.objects.ocpn.obj import OCPetriNet

# object type for binding places used to redstric the OCPN to firing one input and output marker group per firing of the activity
BINDING_OBJECT_TYPE = "_binding"

# prefix for silent transitions in the OCPN
SILENT_TRANSITION_PREFIX = "_silent"


def apply(occn: OCCausalNet, parameters=None) -> OCPetriNet:
    """
    Converts an Object-centric Causal Net (OCCN) to an Object-centric Petri Net (OCPN) with empty markings.
    All cardinalities c!=(1,1) are treated as c=(0, *). Keys for input marker groups are ignored.

    Parameters
    ----------
    occn: OCCausalNet
        The Object-centric Causal Net to convert. Start and end activities must be labeled "START_{object_type}" and "END_{object_type}" respectively.
    parameters: dict, optional
        Additional parameters for the conversion (not used in this implementation).

    Returns
    -------
    OCPetriNet
        The converted Object-centric Petri Net with empty markings.
    """
    activities = occn.activities
    object_types = occn.object_types
    img = occn.input_marker_groups
    omg = occn.output_marker_groups

    assert (
        BINDING_OBJECT_TYPE not in object_types
    ), f"The binding object type {BINDING_OBJECT_TYPE} must not be present in the OCCN."

    assert not (
        any(activity.startswith(SILENT_TRANSITION_PREFIX) for activity in activities)
    ), f"The OCCN must not contain any activities with prefix {SILENT_TRANSITION_PREFIX}."
    
    assert (
        all(f"START_{object_type}" in activities for object_type in object_types)
    ), "All object types must have a corresponding start activity labeled 'START_{object_type}'."
    
    assert (
        all(f"END_{object_type}" in activities for object_type in object_types)
    ), "All object types must have a corresponding end activity labeled 'END_{object_type}'."

    ordered_key_groups_input = build_ordered_key_groups(activities, img)
    ordered_key_groups_output = build_ordered_key_groups(activities, omg)

    places = set()
    place_names = set()
    transitions = set()
    arcs = set()
    
    # create global input binding place
    p_input_binding = OCPetriNet.Place(
        name=f"p{BINDING_OBJECT_TYPE}_global_input",
        object_type=BINDING_OBJECT_TYPE,
    )
    
    places.add(p_input_binding)

    # recursively create OCPN for every activity
    for activity in activities:
        transform_activity(
            activity=activity,
            places=places,
            places_names=place_names,
            transitions=transitions,
            arcs=arcs,
            input_marker_groups=ordered_key_groups_input.get(activity, []),
            output_marker_groups=ordered_key_groups_output.get(activity, []),
            object_types=object_types,
            p_input_binding=p_input_binding,
        )
        
    # create OCPN
    ocpn = OCPetriNet(
        places=places,
        transitions=transitions,
        arcs=arcs,
    )
    
    return ocpn
        


def build_ordered_key_groups(activities, marker_groups):
    """
    Builds a hierarchy of object type groups, key groups and markers for each activity in the given list of activities.
    Object type groups are ordered by object type name.
    Key groups are ordered by the number of markers and by the number of markers with min_count=1 and max_count=1.
    Markers in key groups are ordered by min_count and max_count, effectively putting markers with (1,1) first.

    Parameters
    ----------
    activities: list
        List of activities for which to build the key groups.
    marker_groups: dict
        Dictionary mapping activities to their corresponding marker groups.
        The keys are activity names and the values are lists of marker groups.

    Returns
    -------
    dict
        A dictionary where keys are activities and values are lists of lists of tuples (object_type, object_type_group),
        where each object_type_group is a sorted list of tuples (marker_key, marker_key_group),
        where each marker_key_group is a sorted list of markers.
    """
    # build the hiearchy of key groups, object type groups and ks for each marker group
    key_groups = {}
    for a in activities:
        key_groups[a] = []
        # key_groups[a] is a list of marker groups (dicts)
        # key: values of a marker group is object type: object type group
        # key: values of object type groups are marker_key: key group
        for marker_group in marker_groups.get(a, []):
            marker_group_dict = {}
            for marker in marker_group.markers:
                # add the marker to the key group
                marker_key = marker.marker_key
                object_type = marker.object_type
                if object_type not in marker_group_dict:
                    marker_group_dict[object_type] = {}
                if marker_key not in marker_group_dict[object_type]:
                    marker_group_dict[object_type][marker_key] = []
                marker_group_dict[object_type][marker_key].append(marker)
            if marker_group_dict:
                key_groups[a].append(marker_group_dict)
            

    # sort the key groups
    for a in activities:
        for i, marker_group in enumerate(key_groups[a]):
            for object_type in marker_group:
                for marker_key in marker_group[object_type]:
                    # sort markers by c=(1,1) and then by min_count and max_count
                    marker_group[object_type][marker_key].sort(
                        key=lambda m: (m.min_count != 1, m.max_count != 1, m.min_count, m.max_count)
                    )
                # sort key groups where key groups are ordered by the number of markers.
                # Of the same number, key groups with more markers with c=(1,1) come first,
                # effectively putting key groups with size 1 with 1 marker with c=(1,1) first
                marker_group[object_type] = sorted(
                    marker_group[object_type].items(),
                    key=lambda x: (
                        len(x[1]),  # fewer markers first
                        -sum(
                            1 for m in x[1] if m.min_count == 1 and m.max_count == 1
                        ),  # more (1,1) markers first
                    ),
                )
            # sort the object type groups by alphabetical order of object types
            marker_group = sorted(
                marker_group.items(), key=lambda x: x[0]  # sort by object type name
            )
            key_groups[a][i] = marker_group
        # marker groups for an activity do not need to be sorted
    return key_groups


def transform_activity(
    places,
    places_names,
    transitions,
    arcs,
    activity,
    input_marker_groups,
    output_marker_groups,
    object_types,
    p_input_binding
):
    """
    Creates the necessary places, transitions, and arcs for an activity in the OCPN.

    Parameters
    ----------
    places: set
        Set of places in the OCPN.
    places_names: set
        Set of place names in the OCPN to avoid duplications.
    transitions: set
        Set of transitions in the OCPN.
    arcs: set
        Set of arcs in the OCPN.
    activity: str
        The name of the activity to transform.
    input_marker_groups: list
        List of input marker groups for the activity, where each marker group is a list of object type groups.
    output_marker_groups: list
        List of output marker groups for the activity, where each marker group is a list of object type groups.
    object_types: set
        Set of all object types in the OCPN.
    p_input_binding: OCPetriNet.Place
        The input binding place for the activity construct.
    """
    is_start = activity.startswith("START_")
    if is_start:
        assert not input_marker_groups, f"Start activity {activity} is a START activity and must not have input marker groups."
    is_end = activity.startswith("END_")
    if is_end:
        assert not output_marker_groups, f"End activity {activity} is an END activity and must not have output marker groups."

    # transition
    t = OCPetriNet.Transition(name=activity, label=activity)
    transitions.add(t)

    
    # input and output places
    non_variable_object_types, variable_object_types = get_activity_object_types(
        object_types, input_marker_groups, output_marker_groups
    )
    
        
    for object_type in non_variable_object_types | variable_object_types:
        is_variable = object_type in variable_object_types
        
        p_i = add_place(
            OCPetriNet.Place(
                name=label_activity_place(activity, True, object_type),
                object_type=object_type,
                ),
            places, 
            places_names
        )
        p_o = add_place(
            OCPetriNet.Place(
                name=label_activity_place(activity, False, object_type),
                object_type=object_type,
                ),
            places, 
            places_names
        )
        
        # arcs for input and output places
        add_arc(
            OCPetriNet.Arc(
                source=p_i,
                target=t,
                object_type=object_type,
                is_variable=is_variable,
                ),
            arcs
        )
        add_arc(
            OCPetriNet.Arc(
                source=t,
                target=p_o,
                object_type=object_type,
                is_variable=is_variable,
                ),
            arcs
        )
        
    # binding places
    if not is_start:
        p_i_t = add_place(
            OCPetriNet.Place(
                name=f"p{BINDING_OBJECT_TYPE}#{activity}_input",
                object_type=BINDING_OBJECT_TYPE
                ),
            places,
            places_names,
        )
    if not is_end:
        p_o_t = add_place(
            OCPetriNet.Place(
                name=f"p{BINDING_OBJECT_TYPE}#{activity}_output",
                object_type=BINDING_OBJECT_TYPE
                ),
            places,
            places_names,
        )
    
    # arcs for binding places
    if is_start:
        add_arc(
            OCPetriNet.Arc(
                source=p_input_binding,
                target=t,
                object_type=BINDING_OBJECT_TYPE,    
            ),
            arcs
        )
    else:
        add_arc(
            OCPetriNet.Arc(
                source=p_i_t,
                target=t,
                object_type=BINDING_OBJECT_TYPE,    
            ),
            arcs
        )
    if is_end:
        add_arc(
            OCPetriNet.Arc(
                source=t,
                target=p_input_binding,
                object_type=BINDING_OBJECT_TYPE,    
            ),
            arcs
        )
    else:
        add_arc(
            OCPetriNet.Arc(
                source=t,
                target=p_o_t,
                object_type=BINDING_OBJECT_TYPE,    
            ),
            arcs
        )
    
    # add marker groups
    for marker_group in input_marker_groups:
        transform_marker_group(
            activity=activity,
            places=places,
            places_names=places_names,
            transitions=transitions,
            arcs=arcs,
            marker_group=marker_group,
            p_input_binding=p_input_binding,
            p_output_binding=p_i_t,
            is_output_marker_group=False,  
        )
    for marker_group in output_marker_groups:
        transform_marker_group(
            activity=activity,
            places=places,
            places_names=places_names,
            transitions=transitions,
            arcs=arcs,
            marker_group=marker_group,
            p_input_binding=p_o_t,
            p_output_binding=p_input_binding,
            is_output_marker_group=True,  
        )
    
    


def get_activity_object_types(object_types, input_marker_groups, output_marker_groups):
    """
    Returns a set of non-variable object types and a set of variable object types for the given marker groups.
    Non-variable object types are those where exactly one object of that type is involved in every firing of the activity.

    Parameters
    ----------
    object_types: set
        Set of all object types in the OCPN.
    input_marker_groups: list
        List of input marker groups for the activity, where each marker group is a list of object type groups.
    output_marker_groups: list
        List of output marker groups for the activity, where each marker group is a list of object type groups.

    Returns
    -------
    tuple
        A tuple containing two sets:
        - The first set contains non-variable object types.
        - The second set contains variable object types.
    """
    found_object_types = set()
    # all object types are non-variable unless found otherwise
    non_variable_object_types = {ot: True for ot in object_types}

    for marker_group in input_marker_groups + output_marker_groups:
        for object_type in object_types:
            # Find object type group in the marker_group 
            # (needs to be done like this to catch object types only present in input or output markers)
            matches = [otg for otg in marker_group if otg[0] == object_type]
            if len(matches) > 1:
                raise ValueError(f"More than one object type group found for object type {object_type} in marker_group.")
            object_type_group = matches[0] if matches else (object_type, [])
            
            object_type = object_type_group[0]
            if matches:
                found_object_types.add(object_type)
            markers = [m for key_group in object_type_group[1] for m in key_group[1]]
            # non-variable object types have exactly one marker in its object type group and its cardinalities are (1,1)
            if not (
                len(markers) == 1
                and markers[0].min_count == 1
                and markers[0].max_count == 1
            ):
                # not a non-variable object type
                non_variable_object_types[object_type] = False

    variable_object_types = set(
        ot for ot in found_object_types if not non_variable_object_types[ot]
    )
    non_variable_object_types = set(
        ot for ot in found_object_types if non_variable_object_types[ot]
    )

    return non_variable_object_types, variable_object_types


def transform_marker_group(
    activity,
    places,
    places_names,
    transitions,
    arcs,
    marker_group,
    p_input_binding,
    p_output_binding,
    is_output_marker_group,
):
    """
    Transforms a marker group into a construct of places, transitions, and arcs in the OCPN.
    Places, transitions, and arcs are updated to include the marker group.

    Parameters
    ----------
    activity: str
        The name of the activity which the marker group belongs to.
    places: set
        Set of places in the OCPN.
    places_names: set
        Set of place names in the OCPN to avoid duplications.
    transitions: set
        Set of transitions in the OCPN.
    arcs: set
        Set of arcs in the OCPN.
    marker_group: list
        The marker group to transform, which is a list of object type groups (tuples of object type and list of key groups).
    p_input_binding: OCPetriNet.Place
        The input binding place for the marker group construct.
    p_output_binding: OCPetriNet.Place
        The output binding place for the marker group construct.
    is_output_marker_group: bool
        Indicates if the marker group is an output marker group (True) or an input marker group (False).
    """
    silent_id = get_next_id()
    p_b_i = p_input_binding
    p_b_o = (
        p_output_binding
        if len(marker_group) == 1
        else add_place(
            OCPetriNet.Place(
                name=f"p{BINDING_OBJECT_TYPE}#{silent_id}_1",
                object_type=BINDING_OBJECT_TYPE,
            ),
            places,
            places_names,
        )
    )
    for i, object_type_group in enumerate(marker_group):
        # create construct

        if is_output_marker_group:
            transform_output_object_type_group(
                activity=activity,
                places=places,
                places_names=places_names,
                transitions=transitions,
                arcs=arcs,
                object_type_group=object_type_group,
                p_input_binding=p_b_i,
                p_output_binding=p_b_o,
            )
        else:
            transform_input_object_type_group(
                activity=activity,
                places=places,
                places_names=places_names,
                transitions=transitions,
                arcs=arcs,
                object_type_group=object_type_group,
                p_input_binding=p_b_i,
                p_output_binding=p_b_o,
            )

        # next input and output binding places
        p_b_i = p_b_o
        # for last object type group, p_output_binding is used. We check for >= in case len is 1.
        p_b_o = (
            p_output_binding
            if (i >= len(marker_group) - 2)
            else add_place(
                OCPetriNet.Place(
                    name=f"p{BINDING_OBJECT_TYPE}#{silent_id}_{i + 2}",
                    object_type=BINDING_OBJECT_TYPE,
                ),
                places,
                places_names,
            )
        )


def transform_input_object_type_group(
    activity,
    places,
    places_names,
    transitions,
    arcs,
    object_type_group,
    p_input_binding,
    p_output_binding,
):
    """
    Transforms an object type group of an input marker group into a construct of places, transitions, and arcs in the OCPN.
    Places, transitions, and arcs are updated to include the object type group.
    For input marker groups, we treat all keys as unique.

    Parameters
    ----------
    activity: str
        The name of the activity which the object type group belongs to.
    places: set
        Set of places in the OCPN.
    places_names: set
        Set of place names in the OCPN to avoid duplications.
    transitions: set
        Set of transitions in the OCPN.
    arcs: set
        Set of arcs in the OCPN.
    object_type_group: (object_type, list)
        The object type group to transform, which is a tuple of the object type and a sorted list of key groups.
    p_input_binding: OCPetriNet.Place
        The input binding place for the object type group construct.
    p_output_binding: OCPetriNet.Place
        The output binding place for the object type group construct.
    """
    silent_id = get_next_id()
    # we skip the key groups for input marker groups, as we ignore input marker keys
    # grab all markers of the object type group
    markers = [marker for key_group in object_type_group[1] for marker in key_group[1]]
    # sort them by by c=(1,1) and then by min_count and max_count
    markers.sort(key=lambda m: (m.min_count != 1, m.max_count != 1, m.min_count, m.max_count))

    # proceed with the construct
    ot = object_type_group[0]
    p_b_i = p_input_binding
    p_b_o = (
        p_output_binding
        if len(markers) == 1
        else add_place(
            OCPetriNet.Place(
                name=f"p{BINDING_OBJECT_TYPE}#{silent_id}_1",
                object_type=BINDING_OBJECT_TYPE,
            ),
            places,
            places_names,
        )
    )
    for i, marker in enumerate(markers):
        # transform the marker
        # input and output place
        pi = add_place(
            OCPetriNet.Place(
                name=label_arc_place(marker.related_activity, activity, ot),
                object_type=ot,
            ),
            places,
            places_names,
        )
        po = add_place(
            OCPetriNet.Place(
                name=label_activity_place(activity, True, ot), object_type=ot
            ),
            places,
            places_names,
        )
        # create construct
        transform_marker(
            places=places,
            places_names=places_names,
            transitions=transitions,
            arcs=arcs,
            marker=marker,
            p_input=pi,
            p_output=po,
            p_input_binding=p_b_i,
            p_output_binding=p_b_o,
            is_output_marker=False,
            key_group_length=1,  # does not matter for input markers
            is_last_key_group=True,  # does not matter for input markers
            is_first_marker=i==0
        )

        # next input and output binding places
        p_b_i = p_b_o
        # for last key group, p_output_binding is used. We check for >= in case len is 1.
        p_b_o = (
            p_output_binding
            if (i >= len(markers) - 2)
            else add_place(
                OCPetriNet.Place(
                    name=f"p{BINDING_OBJECT_TYPE}#{silent_id}_{i + 2}",
                    object_type=BINDING_OBJECT_TYPE,
                ),
                places,
                places_names,
            )
        )


def transform_output_object_type_group(
    activity,
    places,
    places_names,
    transitions,
    arcs,
    object_type_group,
    p_input_binding,
    p_output_binding,
):
    """
    Transforms an object type group of an output marker group into a construct of places, transitions, and arcs in the OCPN.
    Places, transitions, and arcs are updated to include the object type group.

    Parameters
    ----------
    activity: str
        The name of the activity which the object type group belongs to.
    places: set
        Set of places in the OCPN.
    places_names: set
        Set of place names in the OCPN to avoid duplications.
    transitions: set
        Set of transitions in the OCPN.
    arcs: set
        Set of arcs in the OCPN.
    object_type_group: (object_type, list)
        The object type group to transform, which is a tuple of the object type and a sorted list of key groups.
    p_input_binding: OCPetriNet.Place
        The input binding place for the object type group construct.
    p_output_binding: OCPetriNet.Place
        The output binding place for the object type group construct.
    """
    silent_id = get_next_id()
    p_b_i = p_input_binding
    p_b_o = (
        p_output_binding
        if len(object_type_group[1]) == 1
        else add_place(
            OCPetriNet.Place(
                name=f"p{BINDING_OBJECT_TYPE}#{silent_id}_1",
                object_type=BINDING_OBJECT_TYPE,
            ),
            places,
            places_names,
        )
    )
    for i, key_group in enumerate(object_type_group[1]):
        # transform the key group
        is_last_key_group = i == len(object_type_group[1]) - 1
        transform_output_key_group(
            activity=activity,
            places=places,
            places_names=places_names,
            transitions=transitions,
            arcs=arcs,
            key_group=key_group,
            p_input_binding=p_b_i,
            p_output_binding=p_b_o,
            is_last_key_group=is_last_key_group,
        )

        # next input and output binding places
        p_b_i = p_b_o
        # for last key group, p_output_binding is used. We check for >= in case len is 1.
        p_b_o = (
            p_output_binding
            if (i >= len(object_type_group[1]) - 2)
            else add_place(
                OCPetriNet.Place(
                    name=f"p{BINDING_OBJECT_TYPE}#{silent_id}_{i + 2}",
                    object_type=BINDING_OBJECT_TYPE,
                ),
                places,
                places_names,
            )
        )


def transform_output_key_group(
    activity,
    places,
    places_names,
    transitions,
    arcs,
    key_group,
    p_input_binding,
    p_output_binding,
    is_last_key_group,
    p_input=None,
):
    """
    Transforms a key group of an output marker group into a construct of places, transitions, and arcs in the OCPN.
    Places, transitions, and arcs are updated to include the key group.

    Parameters
    ----------
    activity: str
        The name of the activity which the key group belongs to.
    places: set
        Set of places in the OCPN.
    places_names: set
        Set of place names in the OCPN to avoid duplications.
    transitions: set
        Set of transitions in the OCPN.
    arcs: set
        Set of arcs in the OCPN.
    key_group: (marker_key, list)
        The key group to transform, which is a tupel of the marker key and a sorted list of markers.
    p_input_binding: OCPetriNet.Place
        The input binding place for the key group construct.
    p_output_binding: OCPetriNet.Place
        The output binding place for the key group construct.
    is_last_key_group: bool
        Indicates if the key group is the last key group of its object type group.
    p_input: OCPetriNet.Place, optional
        The input place for case 2 of this construct. Only to be set when recursively calling this function from case 3.
    """
    ot = key_group[1][0].object_type
    # 3 cases can occur
    if len(key_group[1]) == 1:
        # case 1: key group with one marker
        marker = key_group[1][0]
        # input and output place
        pi = add_place(
            OCPetriNet.Place(
                name=label_activity_place(activity, False, ot), object_type=ot
            ),
            places,
            places_names,
        )
        po = add_place(
            OCPetriNet.Place(
                name=label_arc_place(activity, marker.related_activity, ot),
                object_type=ot,
            ),
            places,
            places_names,
        )
        # create construct
        transform_marker(
            places=places,
            places_names=places_names,
            transitions=transitions,
            arcs=arcs,
            marker=marker,
            p_input=pi,
            p_output=po,
            p_input_binding=p_input_binding,
            p_output_binding=p_output_binding,
            is_output_marker=True,
            key_group_length=len(key_group[1]),
            is_last_key_group=is_last_key_group,
            is_first_marker=True 
         )
    else:
        silent_id = get_next_id()
        if is_last_key_group:
            # case 2: key group with multiple markers, last key group of its object type group
            p_b_i = p_input_binding
            p_b_o = add_place(
                OCPetriNet.Place(
                    name=f"p{BINDING_OBJECT_TYPE}#{silent_id}_1",
                    object_type=BINDING_OBJECT_TYPE,
                ),
                places,
                places_names,
            )

            for i, marker in enumerate(key_group[1]):
                # input and output place
                pi = (
                    p_input
                    if p_input is not None
                    else add_place(  # p_input is set when recursively calling this function from case 3
                        OCPetriNet.Place(
                            name=label_activity_place(activity, False, ot),
                            object_type=ot,
                        ),
                        places,
                        places_names,
                    )
                )
                po = add_place(
                    OCPetriNet.Place(
                        name=label_arc_place(activity, marker.related_activity, ot),
                        object_type=ot,
                    ),
                    places,
                    places_names,
                )
                # create construct
                transform_marker(
                    places=places,
                    places_names=places_names,
                    transitions=transitions,
                    arcs=arcs,
                    marker=marker,
                    p_input=pi,
                    p_output=po,
                    p_input_binding=p_b_i,
                    p_output_binding=p_b_o,
                    is_output_marker=True,
                    key_group_length=len(key_group[1]),
                    is_last_key_group=is_last_key_group,
                    is_first_marker=i == 0,
                )

                # next input and output binding places
                p_b_i = p_b_o
                # for last marker, p_output_binding is used
                p_b_o = (
                    p_output_binding
                    if i >= (len(key_group[1]) - 2)
                    else add_place(
                        OCPetriNet.Place(
                            name=f"p{BINDING_OBJECT_TYPE}#{silent_id}_{i + 2}",
                            object_type=BINDING_OBJECT_TYPE,
                        ),
                        places,
                        places_names,
                    )
                )
        else:
            # case 3: key group with multiple markers, not the last key group of its object type group
            # this is the same as case 2 with some extra on top, so we create the extra and thenn call case 2

            # places
            px = add_place(
                OCPetriNet.Place(
                    name=f"p#{silent_id}_X",
                    object_type=ot,
                ),
                places,
                places_names,
            )
            p_alpha = add_place(
                OCPetriNet.Place(
                    name=f"p{BINDING_OBJECT_TYPE}#{silent_id}_alpha",
                    object_type=BINDING_OBJECT_TYPE,
                ),
                places,
                places_names,
            )
            p_beta = add_place(
                OCPetriNet.Place(
                    name=f"p{BINDING_OBJECT_TYPE}#{silent_id}_beta",
                    object_type=BINDING_OBJECT_TYPE,
                ),
                places,
                places_names,
            )
            # transitions
            t1 = OCPetriNet.Transition(
                name=f"{SILENT_TRANSITION_PREFIX}#{silent_id}_1",
            )
            t2 = OCPetriNet.Transition(
                name=f"{SILENT_TRANSITION_PREFIX}#{silent_id}_2",
            )
            transitions.update([t1, t2])
            # arcs
            p_a_o_ot = add_place(
                OCPetriNet.Place(
                    name=label_activity_place(activity, False, ot),
                    object_type=ot,
                ),
                places,
                places_names,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_a_o_ot,
                    target=t1,
                    object_type=ot,
                    is_variable=True,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t1,
                    target=p_a_o_ot,
                    object_type=ot,
                    is_variable=True,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t1,
                    target=px,
                    object_type=ot,
                    is_variable=True,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_a_o_ot,
                    target=t2,
                    object_type=ot,
                    is_variable=True,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t2,
                    target=px,
                    object_type=ot,
                    is_variable=True,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_input_binding,
                    target=t2,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t2,
                    target=p_alpha,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_alpha,
                    target=t1,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t1,
                    target=p_beta,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )

            # call case 2 by setting is_last_key_group to True, provide px as input place
            transform_output_key_group(
                activity=activity,
                places=places,
                places_names=places_names,
                transitions=transitions,
                arcs=arcs,
                key_group=key_group,
                p_input_binding=p_beta,
                p_output_binding=p_output_binding,
                is_last_key_group=True,
                p_input=px,
            )


def transform_marker(
    places,
    places_names,
    transitions,
    arcs,
    marker,
    p_input,
    p_output,
    p_input_binding,
    p_output_binding,
    is_output_marker,
    key_group_length,
    is_last_key_group,
    is_first_marker
):
    """
    Transforms a marker into a construct of places, transitions, and arcs in the OCPN.
    Places, transitions, and arcs are updated to include the marker.

    Parameters
    ----------
    places: set
        Set of places in the OCPN.
    places_names: set
        Set of place names in the OCPN to which the place name will be added.
    transitions: set
        Set of transitions in the OCPN.
    arcs: set
        Set of arcs in the OCPN.
    marker: Marker
        The marker to transform.
    p_input: OCPetriNet.Place
        The input place for the marker construct.
    p_output: OCPetriNet.Place
        The output place for the marker construct.
    p_input_binding: OCPetriNet.Place
        The input binding place for the marker construct.
    p_output_binding: OCPetriNet.Place
        The output binding place for the marker construct.
    is_output_marker: bool
        True if the marker is part of an output marker group, False if it is part of an input marker group.
    key_group_length: int
        The length of the key group to which the marker belongs. Does not matter for input markers.
    is_last_key_group: bool
        Indicates if the marker is in the last key group of its object type group. Does not matter for input markers.
    is_first_marker: bool
        Indicates if the marker is the first marker in its input object type group. Does not matter for output markers.
    """
    # id to avoid name clashes
    silent_id = get_next_id()

    # 6 cases can occur
    if marker.min_count == 1 and marker.max_count == 1:
        if is_output_marker and not is_last_key_group and key_group_length == 1:
            # case 1: (1,1) marker with duplication
            # places
            for p in [p_input, p_output, p_input_binding, p_output_binding]:
                add_place(p, places, places_names)
            # transitions
            t1 = OCPetriNet.Transition(
                name=f"{SILENT_TRANSITION_PREFIX}#{silent_id}_1",
            )
            t2 = OCPetriNet.Transition(
                name=f"{SILENT_TRANSITION_PREFIX}#{silent_id}_2",
            )
            transitions.update([t1, t2])
            # arcs
            add_arc(
                OCPetriNet.Arc(
                    source=p_input,
                    target=t1,
                    object_type=marker.object_type,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t1,
                    target=p_input,
                    object_type=marker.object_type,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t1,
                    target=p_output,
                    object_type=marker.object_type,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_input,
                    target=t2,
                    object_type=marker.object_type,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t2,
                    target=p_output,
                    object_type=marker.object_type,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_input_binding,
                    target=t1,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_input_binding,
                    target=t2,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t1,
                    target=p_output_binding,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t2,
                    target=p_output_binding,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
        elif (not is_output_marker) and (not is_first_marker):
            # case 2: (1,1) marker with unification
            # places
            for p in [p_input, p_output, p_input_binding, p_output_binding]:
                add_place(p, places, places_names)
            # transitions
            t1 = OCPetriNet.Transition(
                name=f"{SILENT_TRANSITION_PREFIX}#{silent_id}_1",
            )
            t2 = OCPetriNet.Transition(
                name=f"{SILENT_TRANSITION_PREFIX}#{silent_id}_2",
            )
            transitions.update([t1, t2])
            # arcs
            add_arc(
                OCPetriNet.Arc(
                    source=p_input,
                    target=t1,
                    object_type=marker.object_type,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_output,
                    target=t1,
                    object_type=marker.object_type,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t1,
                    target=p_output,
                    object_type=marker.object_type,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_input,
                    target=t2,
                    object_type=marker.object_type,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t2,
                    target=p_output,
                    object_type=marker.object_type,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_input_binding,
                    target=t1,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_input_binding,
                    target=t2,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t1,
                    target=p_output_binding,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t2,
                    target=p_output_binding,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
        else:
            # case 2: (1,1) marker without duplication/unification
            # places
            places.update([p_input, p_output, p_input_binding, p_output_binding])
            # transitions
            t = OCPetriNet.Transition(
                name=f"{SILENT_TRANSITION_PREFIX}#{silent_id}",
            )
            transitions.add(t)
            # arcs
            add_arc(
                OCPetriNet.Arc(
                    source=p_input,
                    target=t,
                    object_type=marker.object_type,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t,
                    target=p_output,
                    object_type=marker.object_type,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_input_binding,
                    target=t,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t,
                    target=p_output_binding,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
    else:
        # marker with min_count != 1 or max_count != 1
        if is_output_marker and not is_last_key_group and key_group_length == 1:
            # case 4: square marker with duplication
            # places
            px = OCPetriNet.Place(
                name=f"p{BINDING_OBJECT_TYPE}#{silent_id}",
                object_type=BINDING_OBJECT_TYPE,
            )
            places.update([p_input, p_output, p_input_binding, p_output_binding, px])
            # transitions
            t1 = OCPetriNet.Transition(
                name=f"{SILENT_TRANSITION_PREFIX}#{silent_id}_1",
            )
            t2 = OCPetriNet.Transition(
                name=f"{SILENT_TRANSITION_PREFIX}#{silent_id}_2",
            )
            transitions.update([t1, t2])
            # arcs
            add_arc(
                OCPetriNet.Arc(
                    source=p_input,
                    target=t1,
                    object_type=marker.object_type,
                    is_variable=True,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t1,
                    target=p_input,
                    object_type=marker.object_type,
                    is_variable=True,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t1,
                    target=p_output,
                    object_type=marker.object_type,
                    is_variable=True,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_input,
                    target=t2,
                    object_type=marker.object_type,
                    is_variable=True,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t2,
                    target=p_output,
                    object_type=marker.object_type,
                    is_variable=True,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_input_binding,
                    target=t2,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t2,
                    target=px,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=px,
                    target=t1,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t1,
                    target=p_output_binding,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
        elif (not is_output_marker) and (not is_first_marker):
            # case 5: square marker with unification 
            # places
            px = OCPetriNet.Place(
                name=f"p{BINDING_OBJECT_TYPE}#{silent_id}",
                object_type=BINDING_OBJECT_TYPE,
            )
            places.update([p_input, p_output, p_input_binding, p_output_binding, px])
            # transitions
            t1 = OCPetriNet.Transition(
                name=f"{SILENT_TRANSITION_PREFIX}#{silent_id}_1",
            )
            t2 = OCPetriNet.Transition(
                name=f"{SILENT_TRANSITION_PREFIX}#{silent_id}_2",
            )
            transitions.update([t1, t2])
            # arcs
            add_arc(
                OCPetriNet.Arc(
                    source=p_input,
                    target=t1,
                    object_type=marker.object_type,
                    is_variable=True,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_output,
                    target=t1,
                    object_type=marker.object_type,
                    is_variable=True,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t1,
                    target=p_output,
                    object_type=marker.object_type,
                    is_variable=True,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_input,
                    target=t2,
                    object_type=marker.object_type,
                    is_variable=True,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t2,
                    target=p_output,
                    object_type=marker.object_type,
                    is_variable=True,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_input_binding,
                    target=t1,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t1,
                    target=px,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=px,
                    target=t2,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t2,
                    target=p_output_binding,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
        else:
            # case 6: square marker without duplication/unification
            # places
            places.update([p_input, p_output, p_input_binding, p_output_binding])
            # transitions
            t = OCPetriNet.Transition(
                name=f"{SILENT_TRANSITION_PREFIX}#{silent_id}",
            )
            transitions.add(t)
            # arcs
            add_arc(
                OCPetriNet.Arc(
                    source=p_input,
                    target=t,
                    object_type=marker.object_type,
                    is_variable=True,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t,
                    target=p_output,
                    object_type=marker.object_type,
                    is_variable=True,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=p_input_binding,
                    target=t,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )
            add_arc(
                OCPetriNet.Arc(
                    source=t,
                    target=p_output_binding,
                    object_type=BINDING_OBJECT_TYPE,
                    is_variable=False,
                ),
                arcs,
            )


def add_arc(arc, arcs):
    """
    Adds an arc to the OCPN and its source and target.

    Parameters
    ----------
    arc: OCPetriNet.Arc
        The arc to add to the OCPN.
    arcs: set
        Set of arcs in the OCPN to which the arc will be added.
    """
    arcs.add(arc)
    arc.source.add_out_arc(arc)
    arc.target.add_in_arc(arc)


def add_place(place, places, places_names) -> OCPetriNet.Place:
    """
    Adds a place to the OCPN and its name to the set of place names.
    Does not add the place if a place with the same name already exists in the OCPN.
    Returns the place if it was added, otherwise returns the existing place.

    Parameters
    ----------
    place: OCPetriNet.Place
        The place to add to the OCPN.
    places: set
        Set of places in the OCPN to which the place will be added.
    places_names: set
        Set of place names in the OCPN to which the place name will be added.

    Returns
    -------
    OCPetriNet.Place
        The place that was added or the existing place if it was not added.
    """
    if place.name not in places_names:
        places.add(place)
        places_names.add(place.name)
        return place
    else:
        # if the place already exists, return the existing place
        for p in places:
            if p.name == place.name:
                return p
        raise ValueError(f"Place with name {place.name} not found in places.")


def label_activity_place(activity_name, is_input, object_type):
    """
    Returns the name for an input / output place of an activity.

    Parameters
    ----------
    activity_name: str
        The name of the activity.
    is_input: bool
        Indicates if the place is an input place (True) or an output place (False).
    object_type: str
        The object type of the activity.

    Returns
    -------
    str
        The name of the place.
    """
    prefix = f"p_{activity_name}_i" if is_input else f"p_{activity_name}_o"
    return f"{prefix}_{object_type}"


def label_arc_place(source, target, object_type):
    """
    Returns the name for the place corresponding to an arc between two activities.

    Parameters
    ----------
    source: str
        The name of the source activity.
    target: str
        The name of the target activity.
    object_type: str
        The object type of the arc.
    """
    return f"p_arc({source},{target})_{object_type}"


def get_next_id():
    """
    Returns a unique id for a silent transition, starting from 0.
    """
    # initialize on first call
    if not hasattr(get_next_id, "counter"):
        get_next_id.counter = 0
    current = get_next_id.counter
    get_next_id.counter += 1
    return current



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from sentence_transformers import SentenceTransformer
from typing import List, Optional, Dict, Any
from enum import Enum
from pm4py.util import exec_utils, constants


class Parameters(Enum):
    EMBEDDING_MODEL = "embedding_model"


def apply(sentence: str, parameters: Optional[Dict[Any, Any]] = None) -> List[float]:
    """
    Encodes a sentence to embeddings

    Parameters
    ----------
    sentence
        Sentence to be encoded
    parameters
        Parameters of the method, including:
        - Parameters.EMBEDDING_MODEL => embedding model to be used (default: all-MiniLM-L6-v2)

    Returns
    -------
    embeddings
        List of numerical embeddings
    """
    if parameters is None:
        parameters = {}

    embedding_model = exec_utils.get_param_value(Parameters.EMBEDDING_MODEL, parameters,
                                                 constants.DEFAULT_EMBEDDING_MODEL)

    model = SentenceTransformer(embedding_model)

    return model.encode([sentence])[0]



"""
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
"""

from collections import Counter, defaultdict
import itertools
from typing import Any, Generic, Set, TypeVar, Union
from copy import deepcopy
from pm4py.objects.oc_causal_net.obj import OCCausalNet


N = TypeVar("N", bound=OCCausalNet)
M = TypeVar("M", bound=OCCausalNet.Marker)
MG = TypeVar("MG", bound=OCCausalNet.MarkerGroup)


class OCCausalNetState(Generic[N], defaultdict):
    """
    The state of an object-centric causal net is a mapping from activities act to
    multisets of outstanding obligations (act2, object_id, object_type) from act2 to act.

    ```
    state = OCCausalNetState({act: Counter([(act2, object_id, object_type)])})
    ```
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initializes the OCCausalNetState, querying unspecified activities defaults to an empty multiset."""
        super().__init__(Counter)
        data_args = args
        if args and args[0] is Counter:
            data_args = args[1:]
        initial_data = dict(*data_args, **kwargs)
        for act, obligations in initial_data.items():
            self[act] = Counter(obligations)

    def __hash__(self):
        return frozenset(
            (act, frozenset(counter.items()))
            for act, counter in self.items()
            if counter
        ).__hash__()

    def __eq__(self, other):
        if not isinstance(other, OCCausalNetState):
            return False
        return all(
            self.get(a, Counter()) == other.get(a, Counter())
            for a in set(self.keys()) | set(other.keys())
        )

    def __le__(self, other):
        for a, self_counter in self.items():
            other_counter = other.get(a, Counter())
            # Every obligation count in self must be less than or equal to the count in other.
            if not all(
                other_counter.get(pred, 0) >= count
                for pred, count in self_counter.items()
            ):
                return False
        return True

    def __add__(self, other):
        result = OCCausalNetState()
        for a, self_counter in self.items():
            result[a] += self_counter
        for a, other_counter in other.items():
            result[a] += other_counter
        return result

    def __sub__(self, other):
        result = OCCausalNetState()
        for a, self_counter in self.items():
            diff = self_counter - other.get(a, Counter())
            if diff != Counter():
                result[a] = diff
        return result

    def __repr__(self):
        # e.g.  [(a, o1[order], a'), ...]
        sorted_entries = sorted(self.items(), key=lambda item: item[0])
        obligations = [
            f"({pred}, {obj_id}[{ot}], {act}):{count}"
            for (act, obl) in sorted_entries
            for ((pred, obj_id, ot), count) in obl.items()
        ]
        return f'[{", ".join(obligations) if obligations else ""}]'

    def __str__(self):
        return self.__repr__()

    def __deepcopy__(self, memodict={}):
        new_state = OCCausalNetState()
        memodict[id(self)] = new_state
        for act, obligations in self.items():
            act_copy = (
                memodict[id(act)] if id(act) in memodict else deepcopy(act, memodict)
            )
            counter_copy = (
                memodict[id(obligations)]
                if id(obligations) in memodict
                else deepcopy(obligations, memodict)
            )
            new_state[act_copy] = counter_copy
        return new_state

    @property
    def activities(self) -> Set:
        """
        Set of activities with outstanding obligations.
        """
        return set(act for act in self.keys() if self[act])


class OCCausalNetSemantics(Generic[N]):
    """
    Class for the semantics of object-centric causal nets
    """

    @classmethod
    def enabled_activities(
        cls,
        occn: N,
        state: OCCausalNetState,
        include_start_activities=False,
        act_to_idx: dict = None,
        ot_to_idx: dict = None,
    ) -> Set:
        """
        Returns the enabled activities in the given state.

        Parameters
        ----------
        occn
            Object-centric causal net
        state
            State of the OCCN
        include_start_activities
            True if start activities should be included in the set.
            Start activities are always enabled.
        act_to_idx
            If activities are denoted in the state by an id instead of their name,
            a dictionary mapping activities to their index has to be provided here.
        ot_to_idx
            If object types are denoted in the state by an id instead of their name,
            a dictionary mapping object types to their index has to be provided here.

        Returns
        -------
        set
            Set of all activities enabled.
        """
        return set(
            act
            for act in occn.activities
            if (include_start_activities or not act.startswith("START_"))
            and cls.is_enabled(occn, act, state, act_to_idx, ot_to_idx)
        )

    @classmethod
    def is_enabled(
        cls,
        occn: N,
        act: str,
        state: OCCausalNetState,
        act_to_idx: dict = None,
        ot_to_idx: dict = None,
    ) -> bool:
        """
        Checks whether a given activity is enabled in a given object-centric
        casal net and state.
        An activity is enabled if there exists an input marker group that can be
        bound. Start activities are always enabled.

        Parameters
        ----------
        occn
            Object-centric causal net
        act
            Activity to check
        state
            State of the OCCN
        act_to_idx
            If activities are denoted in the state by an id instead of their name,
            a dictionary mapping activities to their index has to be provided here.
        ot_to_idx
            If object types are denoted in the state by an id instead of their name,
            a dictionary mapping object types to their index has to be provided here.

        Returns
        -------
        bool
            true if enabled, false otherwise
        """
        # Start activities are always enabled
        if act.startswith("START_"):
            return True

        if act_to_idx:
            act_id = act_to_idx[act]
        else:
            act_id = act
            
        
        # preprocess Counter for this activity to a lookup table where values 
        # are sets of objects with outstanding obligations to this act
        objects = defaultdict(set)
        for rel_act, obj_id, ot_id in state[act_id].keys():
            objects[(rel_act, ot_id)].add(obj_id)
            
        

        imgs = occn.input_marker_groups[act]

        # if there are not outstanding obligations, the activity cannot be enabled
        if not state[act_id]:
            return False

        # check each img
        for img in imgs:
            # markers that allow for consumption of 0 obligations do not need to be enabled
            # at least one marker of the img needs to be enabled
            one_marker_enabled = False
            for marker in img.markers:
                min_count = marker.min_count
                rel_act_id = (
                    marker.related_activity
                    if not act_to_idx
                    else act_to_idx[marker.related_activity]
                )
                ot_id = (
                    marker.object_type
                    if not ot_to_idx
                    else ot_to_idx[marker.object_type]
                )

                num_objects = len(objects[(rel_act_id, ot_id)])
                if num_objects >= max(min_count, 1):
                    one_marker_enabled = True
                elif min_count != 0:
                    # marker is not enabled, move on to next img
                    one_marker_enabled = False
                    break
            if one_marker_enabled:
                return True
        return False

    @classmethod
    def _find_matching_marker_group(
        cls, obligations: dict, marker_groups: list
    ) -> Union[MG, None]:
        """
        Finda a matching marker group that is able to consume/produce the given obligations.

        Parameters
        ----------
        obligations : dict
            Obligations to consume, mapping related activities to a dict mapping
            object types to a set of object ids
        marker_groups : list
            List of marker groups to check

        Returns
        -------
        Union[MG, None]
            A matching marker group if found, None otherwise.
        """
        if not obligations:
            return None
        # Calculate object counts from the obligations
        obj_counts = {
            related_act: {
                ot: len(obligations[related_act][ot]) for ot in obligations[related_act]
            }
            for related_act in obligations
        }

        # check each group
        for mg in marker_groups:
            mg_dict = mg.dict_representation
            
            # Check that markers for all required related activities exist
            # and all required ots are present
            failure = False
            for related_act in obj_counts:
                if related_act not in mg_dict:
                    failure = True
                for ot in obj_counts[related_act]:
                    if ot not in mg_dict[related_act]:
                        failure = True
            if failure:
                continue

            # 1: check that count matches (= is within cardinality bounds)
            counts_match = all(
                mg_dict[related_act][ot][1]
                >= obj_counts.get(related_act, {}).get(ot, 0)
                >= mg_dict[related_act][ot][0]
                for related_act in mg_dict
                for ot in mg_dict[related_act]
            )
            if not counts_match:
                continue

            # 2: check key constraints
            constraints_violated = any(
                obligations.get(rel_act_1, {})
                .get(ot, set())
                .intersection(obligations.get(rel_act_2, {}).get(ot, set()))
                for (rel_act_1, ot, rel_act_2) in mg.key_constraints
            )
            if constraints_violated:
                continue

            # found matching group
            return mg

        return None

    @classmethod
    def is_binding_enabled(
        cls,
        net: N,
        act: str,
        cons: dict[str, dict[str, Set]],
        prod: dict[str, dict[str, Set]],
        state: OCCausalNetState,
    ) -> Union[tuple[MG, MG], None]:
        """
        Checks whether the given binding is enabled in the object-centric causal net.
        A binding is enabled if the activity has input and output marker groupos that
        match the given objects and the state contains all necessary obligations.


        Parameters
        ----------
        net : N
            The object-centric causal net
        act : str
            The activity to bind
        cons : dict[str, dict[str, Set]]
            The obligations to consume, mapping predecessor activities to a dict mapping
            object types to a set of object ids
        prod : dict[str, dict[str, Set]]
            The obligations to produce, mapping successor activities to a dict mapping
            object types to a set of object ids
        state : OCCausalNetState
            The current state of the OCCN

        Returns
        -------
        Union[tuple[MG, MG], None]:
            The input and output marker groups enabling the binding if it is enabled, None otherwise.
        """
        # 1: check that all consumed obligations are present in the state
        if cons:
            if any(
                state[act].get((pred, obj_id, ot), 0) <= 0
                for pred in cons.keys()
                for ot in cons[pred].keys()
                for obj_id in cons[pred][ot]
            ):
                return None
        else:
            if not prod:
                # we need to either consume or produce obligations
                return None

        # 2: Find a matching input marker group
        if act.startswith("START_"):
            if cons:
                return None
            # For START activities, we do not consume obligations, cons has to be empty
            matched_img = None
        else:
            matched_img = cls._find_matching_marker_group(
                cons, net.input_marker_groups[act]
            )
            if matched_img is None:
                return None

        # 3: Find a matching output marker group
        if act.startswith("END_"):
            if prod:
                return None
            # For START activities, we do not consume obligations, prod has to be empty
            matched_omg = None
        else:
            matched_omg = cls._find_matching_marker_group(
                prod, net.output_marker_groups[act]
            )
            if matched_omg is None:
                return None

        return (matched_img, matched_omg)

    @classmethod
    def bind_activity(
        cls,
        net: N,
        act: str,
        cons: dict[str, dict[str, Set]],
        prod: dict[str, dict[str, Set]],
        state: OCCausalNetState,
    ) -> OCCausalNetState:
        """
        Binds an activity in the object-centric causal net.
        For performance reasons, this method does not check whether the binding
        is valid given the current state. If necessary, the caller should
        ensure this, e.g., using the `is_binding_enabled` method.

        Parameters
        ----------
        net : N
            The object-centric causal net
        act : str
            The activity to bind
        cons : dict[str, dict[str, Set]]
            The obligations to consume, mapping predecessor activities to a dict mapping
            object types to a set of object ids
        prod : dict[str, dict[str, Set]]
            The obligations to produce, mapping successor activities to a dict mapping
            object types to a set of object ids
        state : OCCausalNetState
            The current state of the OCCN

        Returns
        -------
        OCCausalNetState
            The new state after binding the activity
        """
        # consume obligations
        if cons:
            consume = OCCausalNetState(
                {
                    act: Counter(
                        [
                            (pred, obj_id, ot)
                            for pred in cons
                            for ot in cons[pred]
                            for obj_id in cons[pred][ot]
                        ]
                    )
                }
            )
            state -= consume

        # produce obligations
        if prod:
            produce = OCCausalNetState(
                {
                    succ: Counter(
                        [
                            (act, obj_id, ot)
                            for ot in prod[succ]
                            for obj_id in prod[succ][ot]
                        ]
                    )
                    for succ in prod
                }
            )
            state += produce

        return state

    @classmethod
    def replay(cls, occn: N, sequence: tuple) -> bool:
        """
        Replays a sequence of bindings on the object-centric causal net.

        Parameters
        ----------
        occn : N
            The object-centric causal net to replay on.
        sequence : tuple
            A sequence of bindings, where each binding is a tuple of (activity_name, consumed_obligations, produced_obligations).
            consumed_obligations and produced_obligations are dictionaries mapping
            related activities to a dict mapping object types to a set of object ids.

        Returns
        -------
        bool
            True if the sequence can be replayed on the net, False otherwise.
        """
        # start in the empty state
        state = OCCausalNetState()

        # replay each binding
        for act, cons, prod in sequence:
            if not cls.is_binding_enabled(occn, act, cons, prod, state):
                return False

            state = cls.bind_activity(occn, act, cons, prod, state)

        # check if we are in the empty state
        if state.activities:
            return False

        return True

    @classmethod
    def enabled_bindings_start_activity(
        cls,
        occn: N,
        act: str,
        object_type: str,
        objects: set,
        act_to_idx: dict = None,
        ot_to_idx: dict = None,
    ) -> tuple:
        """
        Computes all enabled bindings for a start activity with a given set of objects.
        These bindings will produce obligations for at least one of the objects.
        Based on the `enabled_bindings` method, but specialized for start activities.

        Parameters
        ----------
        occn : N
            The object-centric causal net
        act : str
            The start activity to bind
        object_type : str
            The object type of the start activity
        objects : set
            A set of objects to bind to the activity. All bindings will bind at least one.
        act_to_idx : dict, optional
            If activities are denoted in the state by an id instead of their name,
            a dictionary mapping activities to their index has to be provided here.
        ot_to_idx : dict, optional
            If object types are denoted in the state by an id instead of their name,
            a dictionary mapping object types to their index has to be provided here.

        Returns
        -------
        tuple
            A tuple where entries are tuples of activity id, consumed objects, and produced objects.
            Consumed / produces are tuples of (predecessor/successor activity id, objects_per_ot)
            where objects_per_ot is a tuples of entries (object_type, objects)
            where objects is a tuple (obj_id_1, obj_id_2, ...).
            If act_to_idx and ot_to_idx are provided, indices instead of names are
            used for activities and object types.
        """
        assert act.startswith("START_"), "This method is only for start activities."
        if act_to_idx:
            act_id = act_to_idx[act]
        else:
            act_id = act

        if ot_to_idx:
            ot_id = ot_to_idx[object_type]
        else:
            ot_id = object_type

        # create a list of fake consumed tuples that binds the powerset of objects
        fake_consumed = []
        for i in range(len(objects)):
            for combo in itertools.combinations(objects, i + 1):
                fake_consumed.append((
                    (-1, (
                        (ot_id, combo),
                        )
                     )
                    ,))

        # create the corresponding produced tuples
        memo_produced = {}
        memo_ot_assignments = {}

        enabled_bindings = []

        for consumed in fake_consumed:
            possible_produced_for_consumed = cls.__generate_produced_for_consumed(
                occn,
                act,
                consumed,
                memo_produced,
                memo_ot_assignments,
                act_to_idx,
                ot_to_idx,
            )

            for produced in possible_produced_for_consumed:
                enabled_bindings.append((act_id, None, produced))

        return tuple(enabled_bindings)

    @classmethod
    def enabled_bindings(
        cls,
        occn: N,
        act: str,
        state: OCCausalNetState,
        act_to_idx: dict = None,
        ot_to_idx: dict = None,
    ) -> tuple:
        """
        Computes all enabled bindings for a given activity in a given state.

        Parameters
        ----------
        occn : N
            The object-centric causal net
        act : str
            The activity to bind
        state : OCCausalNetState
            The current state of the OCCN
        act_to_idx : dict, optional
            If activities are denoted in the state by an id instead of their name,
            a dictionary mapping activities to their index has to be provided here.
        ot_to_idx : dict, optional
            If object types are denoted in the state by an id instead of their name,
            a dictionary mapping object types to their index has to be provided here.

        Returns
        -------
        tuple
            A tuple where entries are tuples of activity id, consumed objects, and produced objects.
            Consumed / produces are tuples of (predecessor/successor activity id, objects_per_ot)
            where objects_per_ot is a tuples of entries (object_type, objects)
            where objects is a tuple (obj_id_1, obj_id_2, ...).
            If act_to_idx and ot_to_idx are provided, indices instead of names are
            used for activities and object types.
        """
        if act_to_idx:
            act_id = act_to_idx[act]
        else:
            act_id = act

        # outstanding obligations for the activity
        obligations = state[act_id]

        # pre-process obligations to a dict where keys are (related activity, object_type)
        # and values are sets of object ids (neglecting the count)
        obligations_dict = defaultdict(set)
        for (related_act, obj_id, ot_id), _ in obligations.items():
            obligations_dict[(related_act, ot_id)].add(obj_id)

        # ----------- Create all possibilities for consumed -----------

        possible_consumed = cls.__generate_consumed(
            occn, act, obligations_dict, act_to_idx, ot_to_idx
        )
        if not possible_consumed:
            return ()

        # ----------- Create all possible produced tuples per consumed tuple -----------

        final_bindings = []
        # Memoization for produced tuples, keyed by comsumed object sets by object type
        memo_produced = {}
        # Memoization for sub-problem of assigning objects of one ot for one omg
        memo_ot_assignments = {}

        for consumed in possible_consumed:
            # End activities do not produce obligations
            if act.startswith("END_"):
                possible_produced_for_consumed = {None}
            else:
                possible_produced_for_consumed = cls.__generate_produced_for_consumed(
                    occn,
                    act,
                    consumed,
                    memo_produced,
                    memo_ot_assignments,
                    act_to_idx,
                    ot_to_idx,
                )

            # Create final bindings
            for produced in possible_produced_for_consumed:
                final_bindings.append((act_id, consumed, produced))

        return tuple(final_bindings)

    @classmethod
    def __generate_consumed(
        cls,
        occn: N,
        act: str,
        obligations_dict: dict,
        act_to_idx: dict = None,
        ot_to_idx: dict = None,
    ) -> Set[tuple]:
        """
        Generates all possible consumed tuples for a given activity and obligations.
        """
        possible_consumed = set()
        for img in occn.input_marker_groups[act]:
            img_dict = img.dict_representation

            # preprocess key constraints if ids are used
            key_constraints_by_id = []
            if act_to_idx:
                for rel_act_1_id, ot_id, rel_act_2_id in img.key_constraints:
                    key_constraints_by_id.append(
                        (
                            act_to_idx[rel_act_1_id],
                            ot_to_idx[ot_id],
                            act_to_idx[rel_act_2_id],
                        )
                    )
            else:
                key_constraints_by_id = img.key_constraints

            # get all combinations on which objects we can consume given the img
            keys, combinations_iter_list = (
                cls.__generate_predecessor_object_combinations(
                    img_dict, obligations_dict, act_to_idx, ot_to_idx
                )
            )

            if not keys:
                # img not enabled
                continue

            # from the consumed_per_pred, create all combinations of consumed obligations
            # these are added to possible_consumed
            cls.__consumed_from_predecessor_combinations(
                possible_consumed, keys, combinations_iter_list, key_constraints_by_id
            )

        return possible_consumed

    @classmethod
    def __generate_predecessor_object_combinations(
        cls,
        img_dict: dict,
        obligations_dict: dict,
        act_to_idx: dict = None,
        ot_to_idx: dict = None,
    ):
        """
        Generates all combinations of objects from obligations_dict that can be consumed given the img_dict.
        """
        # we build two lists: keys of format (predecessor_id, ot_id) and
        # a list of iterations where each iterator yields all possible object
        # combinations for all (predecessor, ot) pairs
        keys = []
        combinations_iter_list = []

        for predecessor in img_dict:
            for ot in img_dict[predecessor]:
                min_count, max_count = img_dict[predecessor][ot]

                if act_to_idx:
                    pred_id = act_to_idx[predecessor]
                else:
                    pred_id = predecessor

                if ot_to_idx:
                    ot_id = ot_to_idx[ot]
                else:
                    ot_id = ot

                # get all objects for obligations of the predecessor activity and object type
                objects_for_pred = sorted(obligations_dict[(pred_id, ot_id)])

                if len(objects_for_pred) < min_count:
                    # get to next img, this one cannot be bound
                    return [], []
                if not objects_for_pred:
                    # no objects available for this (pred, ot) pair, skip
                    continue
                else:
                    # get all combinations of objects
                    keys.append((pred_id, ot_id))
                    combinations_for_req = itertools.chain.from_iterable(
                        [
                            itertools.combinations(objects_for_pred, r)
                            for r in range(
                                # this (pred, ot) pair may consume min_count to max_count objects
                                min_count,
                                min(max_count, len(objects_for_pred)) + 1,
                            )
                        ]
                    )
                    combinations_iter_list.append(combinations_for_req)

        return keys, combinations_iter_list

    @staticmethod
    def __consumed_from_predecessor_combinations(
        possible_consumed: set,
        keys: list,
        combinations_iter_list: list,
        key_constraints_by_id: list,
    ):
        """
        Generates all consumed tuples given all possible assignments from
        (predecessor, object type) tuples to sets of objects that may be consumed
        by this pair.
        """
        # create cross product of all options per predecessor and object type
        cross_product_iter = itertools.product(*combinations_iter_list)
        for binding_selection in cross_product_iter:
            # create dict mapping predecessor activitiy to (ot_id, objects) tuples
            grouped_by_pred = defaultdict(dict)
            for (pred_id, ot_id), objects in zip(keys, binding_selection):
                if len(objects) > 0:
                    grouped_by_pred[pred_id][ot_id] = objects

            # if all markers of the img are optional, we can skip empty bindings
            if not grouped_by_pred:
                continue

            # Check key constraints
            constraint_violated = False
            for rel_act_1_id, ot_id, rel_act_2_id in key_constraints_by_id:
                objects1 = grouped_by_pred.get(rel_act_1_id, {}).get(ot_id)
                objects2 = grouped_by_pred.get(rel_act_2_id, {}).get(ot_id)
                if objects1 and objects2 and set(objects1).intersection(objects2):
                    # key constraint violated, move on to next binding
                    constraint_violated = True
                    break

            if constraint_violated:
                continue

            # convert to memory-efficient tuple representation
            consumed_tuple = tuple(
                # sorting necessary to ensure duplicate-free tuples in possible_consumed
                (pred_id, tuple(sorted(ot_obj_pairs.items())))
                for pred_id, ot_obj_pairs in sorted(grouped_by_pred.items())
            )
            # add to possible consumed obligations (duplicate-free)
            possible_consumed.add(consumed_tuple)

    @classmethod
    def __generate_produced_for_consumed(
        cls,
        occn: N,
        act: str,
        consumed: tuple,
        memo_produced: dict,
        memo_ot_assignments: dict,
        act_to_idx: dict = None,
        ot_to_idx: dict = None,
    ) -> Set[tuple]:
        """
        Generates all possible produced tuples for a given consumed tuple.
        """
        # 1. Process consumed tuple into sets of consumed objects per ot
        consumed_objects_by_ot = defaultdict(set)
        for _, ot_obj_pairs in consumed:
            for ot_id, objects in ot_obj_pairs:
                consumed_objects_by_ot[ot_id].update(objects)

        # Create hashable key for memo
        consumed_key = cls.__make_hashable(consumed_objects_by_ot)

        # 2. Check memo cache
        if consumed_key in memo_produced:
            possible_produced_for_consumed = memo_produced[consumed_key]
        else:
            # Compute all possible produced tuples
            possible_produced_for_consumed = set()  # set to avoid duplicates
            # Compute per omg; cache to avoid recomputation
            for omg in occn.output_marker_groups[act]:
                produced_for_omg = cls.__generate_produced_for_omg(
                    omg,
                    consumed_objects_by_ot,
                    memo_ot_assignments,
                    act_to_idx,
                    ot_to_idx,
                )
                possible_produced_for_consumed.update(produced_for_omg)

            # Store result in cache
            memo_produced[consumed_key] = possible_produced_for_consumed

        return possible_produced_for_consumed

    @classmethod
    def __generate_produced_for_omg(
        cls, omg, consumed_objects_by_ot, memo, act_to_idx, ot_to_idx
    ):
        """
        Generates all possible produced tuples for a given output marker group and
        consumed objects by object type (these need to be produced).
        """
        omg_dict = omg.dict_representation

        # Pre-process omg requirements and key constraints into efficient lookups in case ids are used
        reqs_by_ot = defaultdict(list)
        key_constraints_by_ot = defaultdict(set)

        for succ_name, ot_map in omg_dict.items():
            succ_id = act_to_idx[succ_name] if act_to_idx else succ_name
            for ot_name, (min_c, max_c) in ot_map.items():
                ot_id = ot_to_idx[ot_name] if ot_to_idx else ot_name
                reqs_by_ot[ot_id].append((succ_id, min_c, max_c))

        for s1_name, ot_name, s2_name in omg.key_constraints:
            s1_id = act_to_idx[s1_name] if act_to_idx else s1_name
            ot_id = ot_to_idx[ot_name] if ot_to_idx else ot_name
            s2_id = act_to_idx[s2_name] if act_to_idx else s2_name
            key_constraints_by_ot[ot_id].add(frozenset([s1_id, s2_id]))

        # Check if objects types match for early exit
        consumed_ots = set(consumed_objects_by_ot.keys())
        required_ots = set(reqs_by_ot.keys())

        if not consumed_ots.issubset(required_ots):
            # omg has no marker for some consumed object type
            return []

        missing_required_ots = required_ots - consumed_ots
        for missing_ot in missing_required_ots:
            # marker has an ot that was not consumed
            # this is an issue if that marker is not optional
            if any(min_c > 0 for _, min_c, _ in reqs_by_ot[missing_ot]):
                return []

        # Get all possible assignments of successor activities to consumed objects
        # this is done per individually per object type
        ot_ids_in_order, assignments_in_order = cls.__generate_successor_assignments(
            omg, memo, consumed_objects_by_ot, reqs_by_ot, key_constraints_by_ot
        )

        # Get cross product of all assignments per object type to get all possible produced tuples
        final_produced_tuples = cls.__produced_from_successor_assignments(
            ot_ids_in_order, assignments_in_order
        )

        return final_produced_tuples

    @staticmethod
    def __generate_successor_assignments(
        omg: MG,
        memo: dict,
        consumed_objects_by_ot: dict,
        reqs_by_ot: dict,
        key_constraints_by_ot: dict,
    ):
        """
        Generates all possible assignments of consumed objects to successor activities.
        """
        # we generate all possible assignments of consumed objects to successor activities
        # and then prune the ones that violate key constraints and
        # the ones that violate the min/max cardinality requirements
        ot_ids_in_order = []
        assignments_in_order = []
        # Solve assignment problem for each object type separetly
        # this is cached since different consumed_objects_by_ot may have
        # the same object sets for the same ot_id
        for ot_id, objects in consumed_objects_by_ot.items():
            memo_key = (id(omg), ot_id, frozenset(objects))
            if memo_key in memo:
                valid_ot_assignments = memo[memo_key]
            else:
                ot_reqs = reqs_by_ot.get(ot_id, [])
                successors_for_ot = [req[0] for req in ot_reqs]
                ot_key_constraints = key_constraints_by_ot.get(ot_id, set())

                per_object_choices = []
                for obj in objects:
                    obj_choices = []
                    # object may be assigned to 1 or more successors
                    for i in range(1, len(successors_for_ot) + 1):
                        # get all combinations of successors of size i
                        for succ_set in itertools.combinations(successors_for_ot, i):
                            # check key constraints
                            is_valid = all(
                                frozenset(pair) not in ot_key_constraints
                                for pair in itertools.combinations(succ_set, 2)
                            )
                            if is_valid:
                                obj_choices.append(succ_set)
                    if not obj_choices:
                        return [], []  # An object has no valid assignment
                    per_object_choices.append(obj_choices)

                # Create all combinations of assignments with cross-product over all objects
                valid_ot_assignments = []
                for assignment_choice in itertools.product(*per_object_choices):
                    final_assignment = defaultdict(list)
                    for obj, succs in zip(objects, assignment_choice):
                        for succ in succs:
                            final_assignment[succ].append(obj)

                    # Check cardinality constraints
                    counts_ok = True
                    for succ_id, min_c, max_c in ot_reqs:
                        count = len(final_assignment.get(succ_id, []))
                        if not (min_c <= count <= max_c):
                            counts_ok = False
                            break

                    if counts_ok:
                        # add to valid assignments
                        canonical_assignment = {
                            s: tuple(sorted(o)) for s, o in final_assignment.items()
                        }
                        valid_ot_assignments.append(canonical_assignment)

                # Store in memoization cache
                memo[memo_key] = valid_ot_assignments

            # Only proceed if we have valid assignments for this ot_id
            if not valid_ot_assignments:
                return [], []

            # Store ot_id and valid assignments
            ot_ids_in_order.append(ot_id)
            assignments_in_order.append(valid_ot_assignments)

        return ot_ids_in_order, assignments_in_order

    @staticmethod
    def __produced_from_successor_assignments(
        ot_ids_in_order: list, assignments_in_order: list
    ) -> Set[tuple]:
        """
        Generates all produced tuples given all possible assignments from
        successor activities to sets of objects that may be produced for this activity.
        """
        final_produced_tuples = set()  # avoid duplicates
        for final_choice in itertools.product(*assignments_in_order):
            # convert to dict first
            produced_grouped_by_succ = defaultdict(dict)

            for ot_id, ot_assignment_dict in zip(ot_ids_in_order, final_choice):
                for succ_id, assigned_objects in ot_assignment_dict.items():
                    if len(assigned_objects) > 0:
                        produced_grouped_by_succ[succ_id][ot_id] = assigned_objects

            # Has to produce at least one object
            if not produced_grouped_by_succ:
                continue

            # Convert to tuple representation
            produced_tuple = tuple(
                (succ_id, tuple(sorted(ot_obj_pairs.items())))
                for succ_id, ot_obj_pairs in sorted(produced_grouped_by_succ.items())
            )
            final_produced_tuples.add(produced_tuple)
        return final_produced_tuples

    @staticmethod
    def __make_hashable(d):
        """Helper function to create a hashable key from a dictionary."""
        # Converts a dict of {ot_id: set(obj_ids)} to a frozenset of items
        # so it can be used as a dictionary key for memoization.
        return frozenset((k, frozenset(v)) for k, v in d.items())



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.simulation.playout.ocpn.variants import extensive
from pm4py.util import exec_utils
from enum import Enum
from pm4py.objects.ocpn.obj import OCPetriNet, OCMarking
from typing import Optional, Dict, Any
from pm4py.objects.ocel.obj import OCEL  


class Variants(Enum):
    EXTENSIVE = extensive


DEFAULT_VARIANT = Variants.EXTENSIVE
VERSIONS = {Variants.EXTENSIVE}


def apply(ocpn: OCPetriNet, initial_marking: OCMarking, final_marking: OCMarking, parameters: Optional[Dict[Any, Any]] = None, variant=DEFAULT_VARIANT) -> OCEL:
    """
    Do the playout of an object-centric Petri net generating an OCEL 2.0.

    Parameters
    -----------
    ocpn
        Object-centric Petri net to play-out
    initial_marking
        Initial marking of the object-centric Petri net
    final_marking
        Final marking of the object-centric Petri net
    parameters
        Parameters of the algorithm
    variant
        Variant of the algorithm to use:
            - Variants.EXTENSIVE: gets all the traces from the model. Can be expensive
            
    Returns
    -----------
    OCEL
        Object-centric event log generated by the playout of the object-centric Petri net
    """
    return exec_utils.get_variant(variant).apply(ocpn, initial_marking, final_marking=final_marking,
                                                 parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Optional, Dict, Any, Tuple
import datetime as dt

import numpy as np
import pandas as pd

from pm4py.objects.ocel.obj import OCEL


def is_null(value) -> bool:
    if value is None:
        return True
    if isinstance(value, (list, dict, tuple, set)):
        return False
    try:
        return bool(pd.isna(value))
    except Exception:
        return False


def normalize_value(value):
    if is_null(value):
        return None
    if isinstance(value, (list, dict, tuple, set)):
        return value
    if isinstance(value, pd.Timestamp):
        return value.isoformat()
    if isinstance(value, (dt.datetime, dt.date)):
        return value.isoformat()
    if isinstance(value, np.datetime64):
        return pd.to_datetime(value).isoformat()
    if isinstance(value, np.timedelta64):
        return str(pd.to_timedelta(value))
    if isinstance(value, np.generic):
        return value.item()
    return value


def get_dataframes_from_ocel(
    ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    if parameters is None:
        parameters = {}

    events = ocel.events.copy()
    for col in events.columns:
        if str(events[col].dtype) == "object":
            events[col] = events[col].map(normalize_value)
        elif "date" in str(events[col].dtype):
            events[col] = events[col].dt.strftime("%Y-%m-%dT%H:%M:%SZ")

    objects = ocel.objects.copy()
    for col in objects.columns:
        if str(objects[col].dtype) == "object":
            objects[col] = objects[col].map(normalize_value)
        elif "date" in str(objects[col].dtype):
            objects[col] = objects[col].dt.strftime("%Y-%m-%dT%H:%M:%SZ")

    return events, objects
